"""
This module contains a base class for translating data produced by ESO systems.

It also provides tools that enable users to automatically add the necessary metadata for compatibility with ESO's
science archive. This is done by querying the archive via programmatic access (pyvo is necessary for this
functionality).

"""

import warnings
from abc import abstractmethod
from datetime import timedelta, datetime, timezone

import astropy.table
import numpy as np
from astropy.io import fits
from astropy.time import Time

try:
    from pyvo.dal import tap
except (ImportError, ModuleNotFoundError):
    tap = None

from ..core.base import Metadatum
from .base import BaseTranslator

ESO_TAP_OBS = "https://archive.eso.org/tap_obs"


class ESOTranslator(BaseTranslator):
    """Abstract class for translators for ESO systems.

    Translators are able to convert non-standard AO telemetry data files into an `AOSystem` object.
    """

    @abstractmethod
    def _get_eso_telescope_name(self) -> str:
        """
        Get value for TELESCOP keyword in ESO's science archive.

        Possible values listed below, according to ESO's Data Interface Control Document.

        --- ESO telescopes ---

        =====================   =============================================================================
        TELESCOP                Telescope
        =====================   =============================================================================
        ESO-NTT                 ESO 3.5-m New Technology Telescope
        ESO-3.6 or ESO-3P6      ESO 3.6-m Telescope
        ESO-VLT-Ui              ESO VLT, Unit Telescope i
        ESO-VLT-Uijkl           ESO VLT, incoherent combination of Unit Telescopes ijkl
        ESO-VLTI-Uijkl          ESO VLT, coherent combination of Unit Telescopes ijkl
        ESO-VLTI-Amnop          ESO VLT, coherent combination of Auxiliary Telescopes mnop
        ESO-VLTI-Uijkl-Amnop    ESO VLT, coherent combination of UTs ijkl and Auxiliary Telescopes mnop
        ESO-VST                 ESO 2.6-m VLT Survey Telescope
        VISTA                   ESO 4-meter Visible and Infrared Telescope for Astronomy
        Sky Monitor             All-Sky Monitor of the Paranal Observatory (MASCOT)
        APEX-12m                Atacama Pathfinder Experiment
        ESO-ELT                 ESO Extremely Large Telescope
        =====================   =============================================================================

        --- Hosted telescopes ---

        =====================   =============================================================================
        TELESCOP                Telescope
        =====================   =============================================================================
        MPI-2.2                 MPI 2.2-m Telescope
        SPECULOOS-<name>        1-m SPECULOOS Telescopes, <name> is one of Galilean moons of Jupiter
        TRAPPIST-S              TRAPPIST South 60-cm Telescope
        APICAM                  ApiCam-3 Fisheye Telescope
        =====================   =============================================================================

        --- Non-ESO telescopes ---

        =====================   =============================================================================
        TELESCOP                Telescope
        =====================   =============================================================================
        UKIRT                   3.6-m United Kingdom Infra-Red Telescope
        WHT                     4.2-m William Herschel Telescope
        =====================   =============================================================================
        """
        pass

    @abstractmethod
    def _get_eso_ao_name(self) -> str:
        """
        Get abbreviation of the adaptive optics system's name, as defined by ESO.
        """
        pass

    @abstractmethod
    def _get_run_id(self) -> str:
        """
        Get the run ID (prog ID) that refers to the data, as defined by ESO.
        """
        pass

    def add_archive_metadata(self, query_archive: bool = False) -> None:
        """
        Adds necessary metadata for ESO Archive to AOSystem.

        Parameters
        ----------
        query_archive : default = False
            Whether the ESO archive should be queried to find relevant metadata already in the archive. Requires pyvo.
        """
        telescope = self._get_eso_telescope_name()
        metadata = {
            'ORIGIN': 'ESO-PARANAL',
            'INSTRUME': self._get_eso_ao_name(),
            'HIERARCH ESO OBS PROG ID': self._get_run_id(),
            'TELESCOP': telescope.replace('%', ''),
            'DATE': datetime.now().isoformat(timespec='milliseconds'),
            'OBJECT': 'AO-TELEM',
            'OBSERVER': 'I, Condor',
            'DATE-OBS': self.system.date_beginning.astimezone(timezone.utc).replace(tzinfo=None).isoformat(
                timespec='milliseconds'),
            'MJD-OBS': Time(self.system.date_beginning, scale='utc').mjd,
            'MJD-END': Time(self.system.date_end, scale='utc').mjd,
            'EXPTIME': (self.system.date_end - self.system.date_beginning).total_seconds(),
            'HIERARCH ESO DPR CATG': 'CALIB',
            'HIERARCH ESO DPR TYPE': f'AO-TELEM,AOT,{self._get_eso_ao_name()}',
            'HIERARCH ESO DPR TECH': self.system.ao_mode,
        }
        if self.system.main_telescope.azimuth is not None:
            metadata['HIERARCH ESO TEL AZ'] = self.system.main_telescope.azimuth
        if self.system.main_telescope.elevation is not None:
            metadata['HIERARCH ESO TEL ALT'] = self.system.main_telescope.elevation

        if query_archive:
            if tap is None:
                raise ImportError("Querying the ESO archive requires the pyvo module."
                                  "You can set the 'query_archive' option to False to skip querying the archive.")
            beg = Time(self.system.date_beginning, scale='utc')

            if '%' in telescope:
                tel_comparator = " LIKE "
            else:
                tel_comparator = "="

            delta = timedelta(hours=1)
            query = f"""
            SELECT TOP 1 ABS(mjd_obs - {beg.mjd}) as diff, *
            FROM dbo.raw
            WHERE mjd_obs between {(beg - delta).mjd} and {(beg + delta).mjd}
                and telescope{tel_comparator}'{telescope}'
                and not dp_type LIKE '%AOT%'
            ORDER BY 1 ASC
            """
            res = tap.TAPService(ESO_TAP_OBS).search(query).to_table()
            if res:
                res = res[0]
                metadata |= {
                    'INSTRUME': res['instrument'],
                    'RA': res['ra'],
                    'DEC': res['dec'],
                    # 'PI-COI': res['pi_coi'],
                    # we do not store PI-COI because it is not needed and can cause issues with special chars
                    'HIERARCH ESO OBS PROG ID': res['prog_id'],
                    'HIERARCH ESO INS MODE': res['ins_mode'],
                }
                if 'HIERARCH ESO TEL AZ' not in metadata:
                    metadata['HIERARCH ESO TEL AZ'] = res['tel_az']
                if 'HIERARCH ESO TEL ALT' not in metadata:
                    metadata['HIERARCH ESO TEL ALT'] = res['tel_alt']
            else:
                warnings.warn(f"Could not find data from telescope '{telescope}' near mjd_obs {beg.mjd} at the "
                              f"ESO Archive")
        self.system.metadata.extend([Metadatum(k, v) for k, v in metadata.items()])

    def get_atmospheric_parameters_from_archive(self) -> astropy.table.Table:
        """
        Get atmospheric data from ESO's Science Archive within the recording period.

        Requires pyvo.
        """
        if tap is None:
            raise ImportError("Querying the ESO archive requires the pyvo module.")

        delta = timedelta(minutes=1)
        beg = (self.system.date_beginning - delta).isoformat(timespec='milliseconds')
        end = (self.system.date_end + delta).isoformat(timespec='milliseconds')

        query = f"""
        SELECT *
        FROM asm.meteo_paranal
        WHERE midpoint_date between '{beg}' and '{end}'
        AND valid=1
        """
        res = tap.TAPService(ESO_TAP_OBS).search(query).to_table()
        return res

    @staticmethod
    def _azimuth_conversion(az: float):
        """
        Convert azimuth from ESO's reference frame to AOT's reference frame.

        ESO's azimuth is measured westwards from the south, while in AOT it is defined as being measured from the
        eastward from the north.

        Parameters
        ----------
        az
            ESO azimuth to be converted
        """
        # We need to subtract the angle between north and south and then apply symmetry.
        return -(az - 180) % 360

    @staticmethod
    def _get_pixel_data_from_table(pix_frame: fits.FITS_rec) -> np.ndarray:
        """
        Get properly reshaped pixel data from FITS binary table data.

        Parameters
        ----------
        pix_frame
            Binary table data containing pixel image.
        """
        sizes_x = pix_frame['WindowSizeX']
        sizes_y = pix_frame['WindowSizeY']
        if np.any(sizes_x != sizes_x[0]) or np.any(sizes_y != sizes_y[0]):
            warnings.warn('Pixel window size seems to change over time.')
        sizes_x = sizes_x[0]
        sizes_y = sizes_y[0]

        return pix_frame['Pixels'][:, :sizes_x * sizes_y].reshape(-1, sizes_x, sizes_y)

    @staticmethod
    def _stack_slopes(data: np.ndarray, slope_axis: int) -> np.ndarray:
        # ESO slopes are ordered tip1, tilt1, tip2, tilt2, etc., so even numbers are tip and odd numbers are tilt.
        # We separate and then stack them.
        if slope_axis == 0:
            tip = data[::2]
            tilt = data[1::2]
        elif slope_axis == 1:
            tip = data[:, ::2]
            tilt = data[:, 1::2]
        else:
            raise NotImplementedError
        return np.stack([tip, tilt], axis=1)



"""
This module contains a class for translating data produced by ESO's ERIS system.
"""

import importlib.resources
import warnings
from datetime import datetime
from pathlib import Path

import numpy as np
from astropy.io import fits

import aotpy
from aotpy.io import image_from_file
from .eso import ESOTranslator


# TODO set image units


class ERISTranslator(ESOTranslator):
    """Contains functions for translating telemetry data produced by ESO's ERIS system.

    Parameters
    ----------
    path
        Path to folder containing all system data.
    """

    def __init__(self, path):
        self._path = Path(path)
        self.system = aotpy.AOSystem(name='ERIS AO')
        self.system.main_telescope = aotpy.MainTelescope(
            uid='ESO VLT UT4',
            enclosing_diameter=8.2,
            inscribed_diameter=8.2
        )
        self.dsm = aotpy.DeformableMirror(
            uid='DSM',
            telescope=self.system.main_telescope,
            n_valid_actuators=1156,
        )
        self.system.wavefront_correctors.append(self.dsm)

        if (ho_loop_file := self._find_bintable_files('hoLoopData')) is not None \
                and (ho_pixel_file := self._find_bintable_files('hoPixelData')) is not None:
            self._handle_ngs_data(ho_loop_file, ho_pixel_file)
        elif (lgs_loop_file := self._find_bintable_files('lgsLoopData')) is not None \
                and (lo_loop_file := self._find_bintable_files('loLoopData')) is not None \
                and (lgs_pixel_file := self._find_bintable_files('lgsPixelData')) is not None \
                and (lo_pixel_file := self._find_bintable_files('loPixelData')) is not None:
            self._handle_lgs_data(lgs_loop_file, lgs_pixel_file, lo_loop_file, lo_pixel_file)
        else:
            raise ValueError('Path does not contain necessary telemetry data.')

    def _find_bintable_files(self, name):
        file = list(self._path.glob(f'{name}_*.fits'))
        if (size := len(file)) == 0:
            return None
        else:
            file = file[0]
            if size > 1:
                warnings.warn(f"Found more than one '{name}' file. Picking '{file.name}'.")
            return file

    def _handle_lgs_data(self, lgs_loop_file, lgs_pixel_file, lo_loop_file, lo_pixel_file):
        self.system.ao_mode = 'LTAO'
        lgs_loop_frame = fits.getdata(lgs_loop_file, extname='LGSLoopFrame')
        lgs_pix_frame = fits.getdata(lgs_pixel_file, extname='LGSPixelFrame')

        lgs_timestamps = lgs_loop_frame['Seconds'] + lgs_loop_frame['USeconds'] / 1.e6
        self.system.date_beginning = datetime.utcfromtimestamp(lgs_timestamps[0])
        self.system.date_end = datetime.utcfromtimestamp(lgs_timestamps[-1])
        lgs_frame_numbers = lgs_loop_frame['FrameCounter']
        lgs_time = aotpy.Time('LGS Loop Time', timestamps=lgs_timestamps.tolist(),
                              frame_numbers=lgs_frame_numbers.tolist())

        active_laser = fits.getheader(self._path / 'JitCtr.CFG.DYNAMIC.fits')['ACTIVE_JITTER']
        llt = aotpy.LaserLaunchTelescope(f'LLT{active_laser}')
        lgs = aotpy.SodiumLaserGuideStar(uid='LGS', laser_launch_telescope=llt)
        self.system.sources.append(lgs)

        eris_data_path = importlib.resources.files('aotpy.data') / 'ERIS'
        with importlib.resources.as_file(eris_data_path / 'ho_subap.fits') as p:
            subaperture_mask = image_from_file(p, name='LGS WFS SUBAPERTURE MASK')
        n_valid_subapertures = np.count_nonzero(subaperture_mask.data != -1)

        reference = self._stack_slopes(fits.getdata(self._path / 'LGSAcq.DET1.REFSLP_WITH_OFFSETS.fits'),
                                       slope_axis=1)[0]
        lgs_wfs = aotpy.ShackHartmann(
            uid='LGS WFS',
            source=lgs,
            n_valid_subapertures=n_valid_subapertures,
            measurements=aotpy.Image('LGS Gradients', self._stack_slopes(lgs_loop_frame['Gradients'], slope_axis=1)),
            ref_measurements=aotpy.Image('LGSAcq.DET1.REFSLP_WITH_OFFSETS', reference),
            subaperture_mask=subaperture_mask,
            mask_offsets=[aotpy.Coordinates(0, 0)],
            subaperture_intensities=aotpy.Image('LGS Intensities', lgs_loop_frame['Intensities'])
        )
        self.system.wavefront_sensors.append(lgs_wfs)

        lgs_wfs.detector = aotpy.Detector(
            uid='LGS DET1',
            dark=image_from_file(self._path / 'LGSAcq.DET1.DARK.fits'),
            weight_map=image_from_file(self._path / 'LGSAcq.DET1.WEIGHT.fits'),
            sky_background=image_from_file(self._path / 'LGSAcq.DET1.BACKGROUND.fits'),
            pixel_intensities=aotpy.Image(name='LGS Pixels',
                                          data=self._get_pixel_data_from_table(lgs_pix_frame),
                                          time=aotpy.Time('LGS Pixel Time',
                                                          frame_numbers=lgs_pix_frame['FrameCounter'].tolist()))
        )
        lgs_wfs.subaperture_size = \
            lgs_wfs.detector.pixel_intensities.data.shape[0] // lgs_wfs.subaperture_mask.data.shape[0]

        s2m = self._stack_slopes(fits.getdata(self._path / 'CLMatrixOptimiser.S2M.fits'), slope_axis=1)
        m2s = self._stack_slopes(fits.getdata(self._path / 'CLMatrixOptimiser.M2S.fits'), slope_axis=0)

        lgs_freq = fits.getheader(self._path / 'LGSDet.CFG.DYNAMIC.fits')['FREQ']
        self.system.loops.append(aotpy.ControlLoop(
            uid='High-order loop',
            input_sensor=lgs_wfs,
            commanded_corrector=self.dsm,
            commands=aotpy.Image('DSM_positions', lgs_loop_frame['DSM_Positions']),
            ref_commands=aotpy.Image('LGSCtr.ACT_POS_REF_MAP_WITH_OFFSETS',
                                     fits.getdata(self._path / 'LGSCtr.ACT_POS_REF_MAP_WITH_OFFSETS.fits')[:, 0]),
            time=lgs_time,
            framerate=lgs_freq,
            time_filter_num=aotpy.Image('LGSCtr.A_TERMS', fits.getdata(self._path / 'LGSCtr.A_TERMS.fits').T),
            time_filter_den=aotpy.Image('LGSCtr.B_TERMS', fits.getdata(self._path / 'LGSCtr.B_TERMS.fits').T),
            measurements_to_modes=aotpy.Image('CLMatrixOptimiser.S2M', s2m),
            modes_to_commands=image_from_file(self._path / 'CLMatrixOptimiser.M2V.fits'),
            commands_to_modes=image_from_file(self._path / 'CLMatrixOptimiser.V2M.fits'),
            modes_to_measurements=aotpy.Image('CLMatrixOptimiser.M2S', m2s),
        ))

        jit = aotpy.TipTiltMirror(
            uid='Jitter',
            telescope=llt
        )

        cm = self._stack_slopes(fits.getdata(self._path / 'JitRecnOptimiser.JitCM.fits'), slope_axis=1)
        im = self._stack_slopes(fits.getdata(self._path / 'JitRecnCalibrat.IM.fits'), slope_axis=0)
        # jit_ref = fits.getdata(path_lgs / 'JitCtr.ACT_POS_REF_MAP_WITH_OFFSETS.fits')[:, 0]
        self.system.loops.append(aotpy.ControlLoop(
            uid='Jitter loop',
            input_sensor=lgs_wfs,
            commanded_corrector=jit,
            commands=aotpy.Image('Jitter_Positions', lgs_loop_frame['Jitter_Positions']),
            # ref_commands=aotpy.Image(f'Jit{i}Ctr.ACT_POS_REF_MAP_WITH_OFFSETS', jit_ref[(i - 1) * 2: i * 2]),
            time=lgs_time,
            framerate=lgs_freq,
            control_matrix=aotpy.Image('JitRecnOptimiser.JitCM', cm),
            interaction_matrix=aotpy.Image('JitRecnCalibrat.IM', im)
        ))

        fsm = aotpy.TipTiltMirror(
            uid='Field Steering Mirror',
            telescope=llt
        )

        # off_ref = fits.getdata(path_lgs / 'JitCtr.OACT_POS_REF_MAP.fits')[0, :]
        # proj_map = fits.getdata(path_lgs / f'JitCtr.PROJ_MAP_SCALED.fits')
        self.system.loops.append(aotpy.OffloadLoop(
            uid='Jitter Offload loop',
            input_corrector=jit,
            commanded_corrector=fsm,
            commands=aotpy.Image('Jitter_Offload', lgs_loop_frame['Jitter_Offload']),
            # ref_commands=aotpy.Image(f'Jit{i}Ctr.OACT_POS_REF_MAP', off_ref[(i - 1) * 2: i * 2]),
            time=lgs_time,
            framerate=lgs_freq,
            # offload_matrix=aotpy.Image(f'Jitter{i}_Offload_Matrix', proj_map[(i - 1) * 2:i * 2, (i - 1) * 2:i * 2])
        ))

        self.system.wavefront_correctors.extend([jit, fsm])

        ngs = aotpy.NaturalGuideStar(uid='NGS')
        self.system.sources.append(ngs)

        lo_loop_frame = fits.getdata(lo_loop_file, extname='LOLoopFrame')
        lo_pix_frame = fits.getdata(lo_pixel_file, extname='LOPixelFrame')

        lo_timestamps = lo_loop_frame['Seconds'] + lo_loop_frame['USeconds'] / 1.e6
        if np.all(lo_timestamps == 0):
            # The file has no timestamps
            lo_timestamps_list = []
        else:
            lo_timestamps_list = lo_timestamps.tolist()
        ho_frame_numbers = lo_loop_frame['HO_FrameCounter']
        lo_time = aotpy.Time('LO Loop Time', timestamps=lo_timestamps_list, frame_numbers=ho_frame_numbers.tolist())

        with importlib.resources.as_file(eris_data_path / 'lo_subap.fits') as p:
            subaperture_mask = image_from_file(p, name='LO WFS SUBAPERTURE MASK')
        n_valid_subapertures = np.count_nonzero(subaperture_mask.data != -1)

        reference = self._stack_slopes(fits.getdata(self._path / 'LOAcq.DET1.REFSLP_WITH_OFFSETS.fits'),
                                       slope_axis=1)[0]
        lo_wfs = aotpy.ShackHartmann(
            uid='LO WFS',
            source=ngs,
            n_valid_subapertures=n_valid_subapertures,
            measurements=aotpy.Image('LO Gradients', self._stack_slopes(lo_loop_frame['Gradients'], slope_axis=1)),
            ref_measurements=aotpy.Image('LOAcq.DET1.REFSLP_WITH_OFFSETS', reference),
            subaperture_mask=subaperture_mask,
            mask_offsets=[aotpy.Coordinates(0, 0)],
            subaperture_intensities=aotpy.Image('LO Intensities', lo_loop_frame['Intensities'])
        )

        lo_pix_fc = lo_pix_frame['FrameCounter']
        aux_fc = np.full_like(lo_pix_fc, -1, dtype=np.int32)
        x = 0
        for i, fc_loop in enumerate(lo_loop_frame['FrameCounter']):
            for j, fc_pix in enumerate(lo_pix_fc[x:]):
                if fc_pix > fc_loop:
                    # we're already past the place
                    x += j
                    break
                if fc_pix == fc_loop:
                    x += j
                    aux_fc[x] = i
                    break
            else:
                break
        where = np.where(aux_fc != -1)
        mask = aux_fc[where]
        masked_lgs = lgs_frame_numbers[mask]
        step = int((masked_lgs[-1] - masked_lgs[0]) / (masked_lgs.size - 1))
        aux_fc[where] = masked_lgs

        first = where[0][0]
        diff = 0
        while True:
            diff += 1
            cur = first - diff
            if cur < 0:
                break
            aux_fc[cur] = masked_lgs[0] - diff * step

        last = where[0][-1]
        diff = 0
        while True:
            diff += 1
            cur = last + diff
            if cur > aux_fc.size - 1:
                break
            aux_fc[cur] = masked_lgs[-1] + diff * step

        lo_wfs.detector = aotpy.Detector(
            uid='LO DET1',
            dark=image_from_file(self._path / 'LOAcq.DET1.DARK.fits'),
            weight_map=image_from_file(self._path / 'LOAcq.DET1.WEIGHT.fits'),
            sky_background=image_from_file(self._path / 'LOAcq.DET1.BACKGROUND.fits'),
            pixel_intensities=aotpy.Image(name='LO Pixels',
                                          data=self._get_pixel_data_from_table(lo_pix_frame),
                                          time=aotpy.Time('LO Pixel Time',
                                                          frame_numbers=aux_fc.tolist()))
        )
        self.system.wavefront_sensors.append(lo_wfs)
        lo_wfs.subaperture_size = \
            lo_wfs.detector.pixel_intensities.data.shape[0] // lo_wfs.subaperture_mask.data.shape[0]

        s2m = self._stack_slopes(fits.getdata(self._path / 'LOCtr.SENSOR_2_MODES.fits'), slope_axis=1)

        self.system.loops.append(aotpy.ControlLoop(
            uid='Low-order loop',
            input_sensor=lo_wfs,
            commanded_corrector=self.dsm,
            commands=aotpy.Image('LO_DSM_positions', lo_loop_frame['LO_DSM_positions']),
            modal_coefficients=aotpy.Image('LO_Modal', lo_loop_frame['LO_Modal']),
            # ref_commands=aotpy.Image('LGSCtr.ACT_POS_REF_MAP_WITH_OFFSETS',
            #                         fits.getdata(path_lgs / 'LGSCtr.ACT_POS_REF_MAP_WITH_OFFSETS.fits')[:, 0]),
            time=lo_time,
            framerate=fits.getheader(self._path / 'LODet.CFG.DYNAMIC.fits')['FREQ'],
            time_filter_num=aotpy.Image('LOCtr.A_TERMS', fits.getdata(self._path / 'LOCtr.A_TERMS.fits').T),
            time_filter_den=aotpy.Image('LOCtr.B_TERMS', fits.getdata(self._path / 'LOCtr.B_TERMS.fits').T),
            measurements_to_modes=aotpy.Image('LOCtr.SENSOR_2_MODES', s2m),
        ))

        trombone = aotpy.LinearStage(
            uid='Trombone',
            telescope=self.system.main_telescope
        )
        self.system.wavefront_correctors.append(trombone)

        s2m = self._stack_slopes(fits.getdata(self._path / 'TruthCtr.SENSOR_2_MODES.fits'), slope_axis=1)
        self.system.loops.append(aotpy.ControlLoop(
            uid='Truth loop',
            input_sensor=lo_wfs,
            commanded_corrector=trombone,
            time_filter_num=aotpy.Image('TruthCtr.A_TERMS', fits.getdata(self._path / 'TruthCtr.A_TERMS.fits').T),
            time_filter_den=aotpy.Image('TruthCtr.B_TERMS', fits.getdata(self._path / 'TruthCtr.B_TERMS.fits').T),
            measurements_to_modes=aotpy.Image('TruthCtr.SENSOR_2_MODES', s2m),
        ))

    def _handle_ngs_data(self, ho_loop_file, ho_pixel_file):
        self.system.ao_mode = 'SCAO'
        ho_loop_frame = fits.getdata(ho_loop_file, extname='HOLoopFrame')
        ho_pix_frame = fits.getdata(ho_pixel_file, extname='HOPixelFrame')

        ho_timestamps = ho_loop_frame['Seconds'] + ho_loop_frame['USeconds'] / 1.e6
        self.system.date_beginning = datetime.utcfromtimestamp(ho_timestamps[0])
        self.system.date_end = datetime.utcfromtimestamp(ho_timestamps[-1])
        ho_frame_numbers = ho_loop_frame['FrameCounter']
        ho_time = aotpy.Time('HO Loop Time', timestamps=ho_timestamps.tolist(),
                             frame_numbers=ho_frame_numbers.tolist())

        ngs = aotpy.NaturalGuideStar('NGS')
        self.system.sources.append(ngs)

        eris_data_path = importlib.resources.files('aotpy.data') / 'ERIS'
        with importlib.resources.as_file(eris_data_path / 'ho_subap.fits') as p:
            subaperture_mask = image_from_file(p, name='LO WFS SUBAPERTURE MASK')
        n_valid_subapertures = np.count_nonzero(subaperture_mask.data != -1)

        reference = self._stack_slopes(fits.getdata(self._path / 'HOAcq.DET1.REFSLP_WITH_OFFSETS.fits'),
                                       slope_axis=1)[0]
        ho_wfs = aotpy.ShackHartmann(
            uid='HO WFS',
            source=ngs,
            n_valid_subapertures=n_valid_subapertures,
            measurements=aotpy.Image('Gradients', self._stack_slopes(ho_loop_frame['Gradients'], slope_axis=1)),
            ref_measurements=aotpy.Image('HOAcq.DET1.REFSLP_WITH_OFFSETS.fits', reference),
            subaperture_mask=subaperture_mask,
            mask_offsets=[aotpy.Coordinates(0, 0)],
            subaperture_intensities=aotpy.Image('Intensities', ho_loop_frame['Intensities'])
        )
        self.system.wavefront_sensors.append(ho_wfs)

        ho_wfs.detector = aotpy.Detector(
            uid='DET1',
            dark=image_from_file(self._path / 'HOAcq.DET1.DARK.fits'),
            weight_map=image_from_file(self._path / 'HOAcq.DET1.WEIGHT.fits'),
            sky_background=image_from_file(self._path / 'HOAcq.DET1.BACKGROUND.fits'),
            pixel_intensities=aotpy.Image(name='HO Pixels',
                                          data=self._get_pixel_data_from_table(ho_pix_frame),
                                          time=aotpy.Time('HO Pixel Time',
                                                          frame_numbers=ho_pix_frame['FrameCounter'].tolist()))
        )
        ho_wfs.subaperture_size = \
            ho_wfs.detector.pixel_intensities.data.shape[0] // ho_wfs.subaperture_mask.data.shape[0]

        s2m = self._stack_slopes(fits.getdata(self._path / 'CLMatrixOptimiser.S2M.fits'), slope_axis=1)
        m2s = self._stack_slopes(fits.getdata(self._path / 'CLMatrixOptimiser.M2S.fits'), slope_axis=0)

        self.system.loops.append(aotpy.ControlLoop(
            uid='High-order loop',
            input_sensor=ho_wfs,
            commanded_corrector=self.dsm,
            commands=aotpy.Image('DSM_positions', ho_loop_frame['DSM_Positions']),
            ref_commands=aotpy.Image('HOCtr.ACT_POS_REF_MAP_WITH_OFFSETS',
                                     fits.getdata(self._path / 'HOCtr.ACT_POS_REF_MAP_WITH_OFFSETS.fits')[:, 0]),
            time=ho_time,
            framerate=fits.getheader(self._path / 'HODet.CFG.DYNAMIC.fits')['FREQ'],
            time_filter_num=aotpy.Image('HOCtr.A_TERMS', fits.getdata(self._path / 'HOCtr.A_TERMS.fits').T),
            time_filter_den=aotpy.Image('HOCtr.B_TERMS', fits.getdata(self._path / 'HOCtr.B_TERMS.fits').T),
            measurements_to_modes=aotpy.Image('CLMatrixOptimiser.S2M', s2m),
            modes_to_commands=image_from_file(self._path / 'CLMatrixOptimiser.M2V.fits'),
            commands_to_modes=image_from_file(self._path / 'CLMatrixOptimiser.V2M.fits'),
            modes_to_measurements=aotpy.Image('CLMatrixOptimiser.M2S', m2s),
        ))

    def _get_eso_telescope_name(self) -> str:
        return 'ESO-VLT-U4'

    def _get_eso_ao_name(self) -> str:
        return 'ERIAO'

    def _get_run_id(self) -> str:
        return '60.A-9278(E)'



"""
This module contains classes and functions that enable writing AOT FITS files.
"""

import numbers
import os

import numpy as np
from astropy.io import fits

import aotpy
from . import _keywords as kw
from .utils import FITSFileImage, FITSURLImage, datetime_to_iso
from ..base import SystemWriter

# NaN is defined here as a single precision float (32-bits), which is the lowest possible float precision in FITS.
# The goal is to ensure that low precision numpy arrays aren't unnecessarily upcasted just because of NaN.
_nan = np.single(np.nan)

# Given that integer values in AOT cannot be negative, we assume the lowest integer in a signed 16-bit integer is null.
# Implicitly, this means aotpy does not support unsigned integers.
_int_min = np.iinfo(np.int16).min


def write_system_to_fits(filename: str, system: aotpy.AOSystem, **kwargs) -> None:
    """
    Write `system` to file specified by `filename` using the FITS format.

    Parameters
    ----------
    filename
        Path to the file that will be written.
    system
        `AOSystem` to be written into a file.
    **kwargs
        Keyword arguments passed on as options to the file handling function.
    """
    FITSWriter(system).write(filename, **kwargs)


class FITSWriter(SystemWriter):
    def __init__(self, system: aotpy.AOSystem) -> None:
        self._system = system

        self._tables: dict[str, dict[
            str, tuple[aotpy.Referenceable, dict[str, numbers.Integral | numbers.Real | list | np.ndarray | str]]]] = {}
        """The outer dictionary converts from table name to the table dictionary. The table dictionary converts from the
        object uid to a tuple. This tuple contains the object itself and a dictionary that converts from the field names
        to the values present in the object."""

        self._images: dict[str, aotpy.Image] = {}

        self._handle_data()

        primary_hdu = self._create_primary_hdu()
        bintable_hdus = self._create_bintable_hdus()
        image_hdus = self._create_image_hdus()
        self._hdus = fits.HDUList([primary_hdu, *bintable_hdus, *image_hdus])

    def write(self, filename: str | os.PathLike, **kwargs) -> None:
        self.get_hdus().writeto(filename, **kwargs)

    def get_hdus(self) -> fits.HDUList:
        """
        Get the list of HDUs that compose the AOT FITS file for the initialized system.
        Returns
        -------
            `HDUList` that composes the AOT FITS file for the initialized system.
        """
        return self._hdus

    def _handle_data(self):
        for atm in self._system.atmosphere_params:
            self._handle_atmospheric_parameters(atm)

        self._handle_telescope(self._system.main_telescope, False)

        for src in self._system.sources:
            self._handle_source(src, False)

        for cam in self._system.scoring_cameras:
            self._handle_scoring_camera(cam)

        for wfs in self._system.wavefront_sensors:
            self._handle_wavefront_sensor(wfs, False)

        for cor in self._system.wavefront_correctors:
            self._handle_wavefront_corrector(cor, False)

        for loop in self._system.loops:
            self._handle_loops(loop)

    def _add_to_table(self, table_name: str, obj: aotpy.Referenceable, row: dict) -> bool:
        fields = kw.TABLE_FIELDS[table_name]
        table = self._tables.setdefault(table_name, {})

        if obj.uid in table:
            if obj is table[obj.uid][0]:
                return False  # already added this exact object, no need to add again
            raise ValueError(f"Repeated value '{obj.uid}' in '{kw.REFERENCE_UID}' column in {table_name}")

        if len(fields) != len(row):
            raise RuntimeError  # This should never happen

        converted = {}
        for field in fields.values():
            value = row[field.name]
            if value is None and field.mandatory:
                raise ValueError(
                    f"'{obj.uid}' missing value in mandatory column '{field.name}' on table '{table_name}'.")

            if field.format == kw.STRING_FORMAT:
                if value is None:
                    value = ''
                elif not isinstance(value, str) and not isinstance(value, np.ndarray) \
                        and np.issubdtype(value.dtype, np.character):
                    raise ValueError(f"Unxpected value in table '{table_name}' column '{field.name}'. "
                                     f"Expected '{field.format}' format, got: {value}.")

            elif field.format == kw.FLOAT_FORMAT:
                if value is None:
                    value = _nan
                elif not isinstance(value, numbers.Real):
                    raise ValueError(f"Unxpected value in table '{table_name}' column '{field.name}'. "
                                     f"Expected '{field.format}' format, got: {value}.")

            elif field.format == kw.INTEGER_FORMAT:
                if value is None:
                    value = _int_min
                elif not isinstance(value, numbers.Integral):
                    raise ValueError(f"Unxpected value in table '{table_name}' column '{field.name}'. "
                                     f"Expected '{field.format}' format, got: {value}.")

            elif field.format == kw.LIST_FORMAT:
                if value is None:
                    value = []
                else:
                    try:
                        value = [_nan if v is None else v for v in value]
                    except TypeError:
                        # Not iterable
                        raise ValueError(f"Unxpected value in table '{table_name}' column '{field.name}'. "
                                         f"Expected '{field.format}' format, got: {value}.") from None
            else:
                raise NotImplementedError

            converted[field.name] = value

        table[obj.uid] = (obj, converted)
        return True

    @staticmethod
    def _create_row_reference(uid: str) -> str:
        return f'{kw.ROW_REFERENCE}<{uid}>'

    def _handle_time(self, time: aotpy.Time) -> str | None:
        if time is None:
            return None
        if time.timestamps is not None and time.frame_numbers is not None \
                and len(time.timestamps) != len(time.frame_numbers):
            raise ValueError(f"Error in Time '{time.uid}': If both 'timestamps' and 'frame_numbers' are non-null, they"
                             f"must have the same length.")
        row = {
            kw.REFERENCE_UID: time.uid,
            kw.TIME_TIMESTAMPS: time.timestamps,
            kw.TIME_FRAME_NUMBERS: time.frame_numbers
        }
        self._add_to_table(kw.TIME_TABLE, time, row)
        return self._create_row_reference(time.uid)

    def _handle_image(self, image: aotpy.Image) -> str | None:
        if image is None:
            return None
        self._handle_time(image.time)
        if isinstance(image, FITSFileImage):
            if image.index is not None:
                return f'{kw.FILE_REFERENCE}<{image.filename}>{image.index}'
            else:
                return f'{kw.FILE_REFERENCE}<{image.filename}>'
        if isinstance(image, FITSURLImage):
            if image.index is not None:
                return f'{kw.URL_REFERENCE}<{image.url}>{image.index}'
            else:
                return f'{kw.URL_REFERENCE}<{image.url}>'
        else:
            reference = f'{kw.INTERNAL_REFERENCE}<{image.name}>'
            if image.name in self._images:
                if image is self._images[image.name]:
                    # Image has already been handled before
                    return reference
                raise ValueError(f'Repeated image name {image.name}.')
            self._images[image.name] = image
            return reference

    def _handle_atmospheric_parameters(self, atm: aotpy.AtmosphericParameters) -> None:
        if atm is None:
            raise ValueError("'AOSystem.atmosphere_params' list cannot contain 'None' items.")

        row = {
            kw.REFERENCE_UID: atm.uid,
            kw.ATMOSPHERIC_PARAMETERS_WAVELENGTH: atm.wavelength,
            kw.TIME_REFERENCE: self._handle_time(atm.time),
            kw.ATMOSPHERIC_PARAMETERS_R0: atm.r0,
            kw.ATMOSPHERIC_PARAMETERS_FWHM: atm.fwhm,
            kw.ATMOSPHERIC_PARAMETERS_TAU0: atm.tau0,
            kw.ATMOSPHERIC_PARAMETERS_THETA0: atm.theta0,
            kw.ATMOSPHERIC_PARAMETERS_LAYERS_WEIGHT: self._handle_image(atm.layers_weight),
            kw.ATMOSPHERIC_PARAMETERS_LAYERS_HEIGHT: self._handle_image(atm.layers_height),
            kw.ATMOSPHERIC_PARAMETERS_LAYERS_L0: self._handle_image(atm.layers_l0),
            kw.ATMOSPHERIC_PARAMETERS_LAYERS_WIND_SPEED: self._handle_image(atm.layers_wind_speed),
            kw.ATMOSPHERIC_PARAMETERS_LAYERS_WIND_DIRECTION: self._handle_image(atm.layers_wind_direction),
            kw.TRANSFORMATION_MATRIX: self._handle_image(atm.transformation_matrix),
        }
        self._add_to_table(kw.ATMOSPHERIC_PARAMETERS_TABLE, atm, row)

    def _handle_aberration(self, abr: aotpy.Aberration) -> str | None:
        if abr is None:
            return None

        row = {
            kw.REFERENCE_UID: abr.uid,
            kw.ABERRATION_MODES: self._handle_image(abr.modes),
            kw.ABERRATION_COEFFICIENTS: self._handle_image(abr.coefficients),
            kw.ABERRATION_X_OFFSETS: [off.x for off in abr.offsets],
            kw.ABERRATION_Y_OFFSETS: [off.y for off in abr.offsets]
        }
        self._add_to_table(kw.ABERRATIONS_TABLE, abr, row)
        return self._create_row_reference(abr.uid)

    def _handle_telescope(self, tel: aotpy.Telescope, referenced: bool, llt: bool = False) -> str | None:
        if tel is None:
            if referenced:
                return None
            raise ValueError("'AOSystem.main_telescope' must not be 'None'.")

        main = False
        if isinstance(tel, aotpy.MainTelescope):
            if llt:
                raise ValueError("Referenced telescope must have type 'LaserLaunchTelescope'.")
            main = True
            tel_type = kw.TELESCOPE_TYPE_MAIN
        else:
            if not referenced:
                raise ValueError("'AOSystem.main_telescope' must have type 'MainTelescope'.")
            if isinstance(tel, aotpy.LaserLaunchTelescope):
                tel_type = kw.TELESCOPE_TYPE_LLT
            else:
                if llt:
                    raise ValueError("Referenced telescope must have type 'LaserLaunchTelescope'.")
                raise NotImplementedError

        if tel.segments is None:
            raise ValueError("Telescope 'segments' variable cannot be 'None'.")
        else:
            if isinstance(tel.segments, aotpy.Monolithic):
                segments_type = kw.TELESCOPE_SEGMENT_TYPE_MONOLITHIC
            elif isinstance(tel.segments, aotpy.HexagonalSegments):
                segments_type = kw.TELESCOPE_SEGMENT_TYPE_HEXAGON
            elif isinstance(tel.segments, aotpy.CircularSegments):
                segments_type = kw.TELESCOPE_SEGMENT_TYPE_CIRCLE
            else:
                raise NotImplementedError
            segments_size = tel.segments.size
            segments_x = [coord.x for coord in tel.segments.coordinates]
            segments_y = [coord.y for coord in tel.segments.coordinates]

        row = {
            kw.REFERENCE_UID: tel.uid,
            kw.TELESCOPE_TYPE: tel_type,
            kw.TELESCOPE_LATITUDE: tel.latitude,
            kw.TELESCOPE_LONGITUDE: tel.longitude,
            kw.TELESCOPE_ELEVATION: tel.elevation,
            kw.TELESCOPE_AZIMUTH: tel.azimuth,
            kw.TELESCOPE_PARALLACTIC: tel.parallactic,
            kw.TELESCOPE_PUPIL_MASK: self._handle_image(tel.pupil_mask),
            kw.TELESCOPE_PUPIL_ANGLE: tel.pupil_angle,
            kw.TELESCOPE_ENCLOSING_D: tel.enclosing_diameter,
            kw.TELESCOPE_INSCRIBED_D: tel.inscribed_diameter,
            kw.TELESCOPE_OBSTRUCTION_D: tel.obstruction_diameter,
            kw.TELESCOPE_SEGMENTS_TYPE: segments_type,
            kw.TELESCOPE_SEGMENTS_SIZE: segments_size,
            kw.TELESCOPE_SEGMENTS_X: segments_x,
            kw.TELESCOPE_SEGMENTS_Y: segments_y,
            kw.TRANSFORMATION_MATRIX: self._handle_image(tel.transformation_matrix),
            kw.ABERRATION_REFERENCE: self._handle_aberration(tel.aberration)
        }
        if self._add_to_table(kw.TELESCOPES_TABLE, tel, row) and referenced and main:
            raise ValueError(
                "Telescope references cannot reference a 'MainTelescope' object that is not AOSystem.main_telescope.")
        return self._create_row_reference(tel.uid)

    def _handle_source(self, src: aotpy.Source, referenced: bool) -> str | None:
        if src is None:
            if referenced:
                return None
            raise ValueError("'AOSystem.sources' list cannot contain 'None' items.")

        if isinstance(src, aotpy.ScienceStar):
            src_type = kw.SOURCE_TYPE_SCIENCE_STAR
        elif isinstance(src, aotpy.NaturalGuideStar):
            src_type = kw.SOURCE_TYPE_NATURAL_GUIDE_STAR
        elif isinstance(src, aotpy.SodiumLaserGuideStar):
            src_type = kw.SOURCE_TYPE_SODIUM_LASER_GUIDE_STAR
            row = {
                kw.REFERENCE_UID: src.uid,
                kw.SOURCE_SODIUM_LGS_HEIGHT: src.height,
                kw.SOURCE_SODIUM_LGS_PROFILE: self._handle_image(src.profile),
                kw.SOURCE_SODIUM_LGS_ALTITUDES: src.altitudes,
                kw.LASER_LAUNCH_TELESCOPE_REFERENCE: self._handle_telescope(src.laser_launch_telescope, True)
            }
            self._add_to_table(kw.SOURCES_SODIUM_LGS_TABLE, src, row)
        elif isinstance(src, aotpy.RayleighLaserGuideStar):
            src_type = kw.SOURCE_TYPE_RAYLEIGH_LASER_GUIDE_STAR
            row = {
                kw.REFERENCE_UID: src.uid,
                kw.SOURCE_RAYLEIGH_LGS_DISTANCE: src.distance,
                kw.SOURCE_RAYLEIGH_LGS_DEPTH: src.depth,
                kw.LASER_LAUNCH_TELESCOPE_REFERENCE: self._handle_telescope(src.laser_launch_telescope, True)
            }
            self._add_to_table(kw.SOURCES_RAYLEIGH_LGS_TABLE, src, row)
        else:
            raise NotImplementedError

        row = {
            kw.REFERENCE_UID: src.uid,
            kw.SOURCE_TYPE: src_type,
            kw.SOURCE_RIGHT_ASCENSION: src.right_ascension,
            kw.SOURCE_DECLINATION: src.declination,
            kw.SOURCE_ELEVATION_OFFSET: src.elevation_offset,
            kw.SOURCE_AZIMUTH_OFFSET: src.azimuth_offset,
            kw.SOURCE_WIDTH: src.width
        }
        if self._add_to_table(kw.SOURCES_TABLE, src, row) and referenced:
            raise ValueError(f"Source '{src.uid}' is referenced but not present in 'AOSystem.sources'")
        return self._create_row_reference(src.uid)

    def _handle_detector(self, det: aotpy.Detector) -> str | None:
        if det is None:
            return None

        row = {
            kw.REFERENCE_UID: det.uid,
            kw.DETECTOR_TYPE: det.type,
            kw.DETECTOR_SAMPLING_TECHNIQUE: det.sampling_technique,
            kw.DETECTOR_SHUTTER_TYPE: det.shutter_type,
            kw.DETECTOR_FLAT_FIELD: self._handle_image(det.flat_field),
            kw.DETECTOR_READOUT_NOISE: det.readout_noise,
            kw.DETECTOR_PIXEL_INTENSITIES: self._handle_image(det.pixel_intensities),
            kw.DETECTOR_INTEGRATION_TIME: det.integration_time,
            kw.DETECTOR_COADDS: det.coadds,
            kw.DETECTOR_DARK: self._handle_image(det.dark),
            kw.DETECTOR_WEIGHT_MAP: self._handle_image(det.weight_map),
            kw.DETECTOR_QUANTUM_EFFICIENCY: det.quantum_efficiency,
            kw.DETECTOR_PIXEL_SCALE: det.pixel_scale,
            kw.DETECTOR_BINNING: det.binning,
            kw.DETECTOR_BANDWIDTH: det.bandwidth,
            kw.DETECTOR_TRANSMISSION_WAVELENGTH: det.transmission_wavelength,
            kw.DETECTOR_TRANSMISSION: det.transmission,
            kw.DETECTOR_SKY_BACKGROUND: self._handle_image(det.sky_background),
            kw.DETECTOR_GAIN: det.gain,
            kw.DETECTOR_EXCESS_NOISE: det.excess_noise,
            kw.DETECTOR_FILTER: det.filter,
            kw.DETECTOR_BAD_PIXEL_MAP: self._handle_image(det.bad_pixel_map),
            kw.DETECTOR_DYNAMIC_RANGE: det.dynamic_range,
            kw.DETECTOR_READOUT_RATE: det.readout_rate,
            kw.DETECTOR_FRAME_RATE: det.frame_rate,
            kw.TRANSFORMATION_MATRIX: self._handle_image(det.transformation_matrix),
        }
        self._add_to_table(kw.DETECTORS_TABLE, det, row)
        return self._create_row_reference(det.uid)

    def _handle_scoring_camera(self, cam: aotpy.ScoringCamera) -> None:
        if cam is None:
            raise ValueError("'AOSystem.scoring_cameras' list cannot contain 'None' items.")

        row = {
            kw.REFERENCE_UID: cam.uid,
            kw.SCORING_CAMERA_PUPIL_MASK: self._handle_image(cam.pupil_mask),
            kw.SCORING_CAMERA_WAVELENGTH: cam.wavelength,
            kw.TRANSFORMATION_MATRIX: self._handle_image(cam.transformation_matrix),
            kw.DETECTOR_REFERENCE: self._handle_detector(cam.detector),
            kw.ABERRATION_REFERENCE: self._handle_aberration(cam.aberration)
        }
        self._add_to_table(kw.SCORING_CAMERAS_TABLE, cam, row)

    def _handle_wavefront_sensor(self, wfs: aotpy.WavefrontSensor, referenced: bool) -> str | None:
        if wfs is None:
            if referenced:
                return None
            raise ValueError("'AOSystem.wavefront_sensors' list cannot contain 'None' items.")

        if isinstance(wfs, aotpy.ShackHartmann):
            wfs_type = kw.WAVEFRONT_SENSOR_TYPE_SHACK_HARTMANN
            row = {
                kw.REFERENCE_UID: wfs.uid,
                kw.WAVEFRONT_SENSOR_SHACK_HARTMANN_CENTROIDING_ALGORITHM: wfs.centroiding_algorithm,
                kw.WAVEFRONT_SENSOR_SHACK_HARTMANN_CENTROID_GAINS: self._handle_image(wfs.centroid_gains),
                kw.WAVEFRONT_SENSOR_SHACK_HARTMANN_SPOT_FWHM: self._handle_image(wfs.spot_fwhm)
            }
            self._add_to_table(kw.WAVEFRONT_SENSORS_SHACK_HARTMANN_TABLE, wfs, row)
        elif isinstance(wfs, aotpy.Pyramid):
            wfs_type = kw.WAVEFRONT_SENSOR_TYPE_PYRAMID
            row = {
                kw.REFERENCE_UID: wfs.uid,
                kw.WAVEFRONT_SENSOR_PYRAMID_N_SIDES: wfs.n_sides,
                kw.WAVEFRONT_SENSOR_PYRAMID_MODULATION: wfs.modulation
            }
            self._add_to_table(kw.WAVEFRONT_SENSORS_PYRAMID_TABLE, wfs, row)
        else:
            raise NotImplementedError

        row = {
            kw.REFERENCE_UID: wfs.uid,
            kw.WAVEFRONT_SENSOR_TYPE: wfs_type,
            kw.SOURCE_REFERENCE: self._handle_source(wfs.source, True),
            kw.WAVEFRONT_SENSOR_DIMENSIONS: wfs.dimensions,
            kw.WAVEFRONT_SENSOR_N_VALID_SUBAPERTURES: wfs.n_valid_subapertures,
            kw.WAVEFRONT_SENSOR_MEASUREMENTS: self._handle_image(wfs.measurements),
            kw.WAVEFRONT_SENSOR_REF_MEASUREMENTS: self._handle_image(wfs.ref_measurements),
            kw.WAVEFRONT_SENSOR_SUBAPERTURE_MASK: self._handle_image(wfs.subaperture_mask),
            kw.WAVEFRONT_SENSOR_MASK_X_OFFSETS: [off.x for off in wfs.mask_offsets],
            kw.WAVEFRONT_SENSOR_MASK_Y_OFFSETS: [off.y for off in wfs.mask_offsets],
            kw.WAVEFRONT_SENSOR_SUBAPERTURE_SIZE: wfs.subaperture_size,
            kw.WAVEFRONT_SENSOR_SUBAPERTURE_INTENSITIES: self._handle_image(wfs.subaperture_intensities),
            kw.WAVEFRONT_SENSOR_WAVELENGTH: wfs.wavelength,
            kw.WAVEFRONT_SENSOR_OPTICAL_GAIN: self._handle_image(wfs.optical_gain),
            kw.TRANSFORMATION_MATRIX: self._handle_image(wfs.transformation_matrix),
            kw.DETECTOR_REFERENCE: self._handle_detector(wfs.detector),
            kw.ABERRATION_REFERENCE: self._handle_aberration(wfs.aberration),
            kw.NCPA_REFERENCE: self._handle_aberration(wfs.non_common_path_aberration)
        }
        if self._add_to_table(kw.WAVEFRONT_SENSORS_TABLE, wfs, row) and referenced:
            raise ValueError(
                f"Wavefront Sensor '{wfs.uid}' is referenced but not present in 'AOSystem.wavefront_sensors'")
        return self._create_row_reference(wfs.uid)

    def _handle_wavefront_corrector(self, cor: aotpy.WavefrontCorrector, referenced: bool) -> str | None:
        if cor is None:
            if referenced:
                return None
            raise ValueError("'AOSystem.wavefront_correctors' list cannot contain 'None' items.")

        if isinstance(cor, aotpy.DeformableMirror):
            cor_type = kw.WAVEFRONT_CORRECTOR_TYPE_DM
            row = {
                kw.REFERENCE_UID: cor.uid,
                kw.WAVEFRONT_CORRECTOR_DM_ACTUATORS_X: [coord.x for coord in cor.actuator_coordinates],
                kw.WAVEFRONT_CORRECTOR_DM_ACTUATORS_Y: [coord.y for coord in cor.actuator_coordinates],
                kw.WAVEFRONT_CORRECTOR_DM_INFLUENCE_FUNCTION: self._handle_image(cor.influence_function),
                kw.WAVEFRONT_CORRECTOR_DM_STROKE: cor.stroke
            }
            self._add_to_table(kw.WAVEFRONT_CORRECTORS_DM_TABLE, cor, row)
        elif isinstance(cor, aotpy.TipTiltMirror):
            cor_type = kw.WAVEFRONT_CORRECTOR_TYPE_TTM
        elif isinstance(cor, aotpy.LinearStage):
            cor_type = kw.WAVEFRONT_CORRECTOR_TYPE_LS
        else:
            raise NotImplementedError
        row = {
            kw.REFERENCE_UID: cor.uid,
            kw.WAVEFRONT_CORRECTOR_TYPE: cor_type,
            kw.TELESCOPE_REFERENCE: self._handle_telescope(cor.telescope, True),
            kw.WAVEFRONT_CORRECTOR_N_VALID_ACTUATORS: cor.n_valid_actuators,
            kw.WAVEFRONT_CORRECTOR_PUPIL_MASK: self._handle_image(cor.pupil_mask),
            kw.WAVEFRONT_CORRECTOR_TFZ_NUM: cor.tfz_num,
            kw.WAVEFRONT_CORRECTOR_TFZ_DEN: cor.tfz_den,
            kw.TRANSFORMATION_MATRIX: self._handle_image(cor.transformation_matrix),
            kw.ABERRATION_REFERENCE: self._handle_aberration(cor.aberration)
        }
        if self._add_to_table(kw.WAVEFRONT_CORRECTORS_TABLE, cor, row) and referenced:
            raise ValueError(
                f"Wavefront Corrector '{cor.uid}' is referenced but not present in 'AOSystem.wavefront_correctors'")
        return self._create_row_reference(cor.uid)

    def _handle_loops(self, loop: aotpy.Loop) -> None:
        if loop is None:
            raise ValueError("'AOSystem.loops' list cannot contain 'None' items.")

        if isinstance(loop, aotpy.ControlLoop):
            loop_type = kw.LOOPS_TYPE_CONTROL
            row = {
                kw.REFERENCE_UID: loop.uid,
                kw.LOOPS_CONTROL_INPUT_SENSOR: self._handle_wavefront_sensor(loop.input_sensor, True),
                kw.LOOPS_CONTROL_MODES: self._handle_image(loop.modes),
                kw.LOOPS_CONTROL_MODAL_COEFFICIENTS: self._handle_image(loop.modal_coefficients),
                kw.LOOPS_CONTROL_CONTROL_MATRIX: self._handle_image(loop.control_matrix),
                kw.LOOPS_CONTROL_MEASUREMENTS_TO_MODES: self._handle_image(loop.measurements_to_modes),
                kw.LOOPS_CONTROL_MODES_TO_COMMANDS: self._handle_image(loop.modes_to_commands),
                kw.LOOPS_CONTROL_INTERACTION_MATRIX: self._handle_image(loop.interaction_matrix),
                kw.LOOPS_CONTROL_COMMANDS_TO_MODES: self._handle_image(loop.commands_to_modes),
                kw.LOOPS_CONTROL_MODES_TO_MEASUREMENTS: self._handle_image(loop.modes_to_measurements),
                kw.LOOPS_CONTROL_RESIDUAL_COMMANDS: self._handle_image(loop.residual_commands)
            }
            self._add_to_table(kw.LOOPS_CONTROL_TABLE, loop, row)
        elif isinstance(loop, aotpy.OffloadLoop):
            loop_type = kw.LOOPS_TYPE_OFFLOAD
            row = {
                kw.REFERENCE_UID: loop.uid,
                kw.LOOPS_OFFLOAD_INPUT_CORRECTOR: self._handle_wavefront_corrector(loop.input_corrector, True),
                kw.LOOPS_OFFLOAD_OFFLOAD_MATRIX: self._handle_image(loop.offload_matrix)
            }
            self._add_to_table(kw.LOOPS_OFFLOAD_TABLE, loop, row)
        else:
            raise NotImplementedError
        row = {
            kw.REFERENCE_UID: loop.uid,
            kw.LOOPS_TYPE: loop_type,
            kw.LOOPS_COMMANDED: self._handle_wavefront_corrector(loop.commanded_corrector, True),
            kw.TIME_REFERENCE: self._handle_time(loop.time),
            kw.LOOPS_STATUS: kw.LOOPS_STATUS_CLOSED if loop.closed else kw.LOOPS_STATUS_OPEN,
            kw.LOOPS_COMMANDS: self._handle_image(loop.commands),
            kw.LOOPS_REF_COMMANDS: self._handle_image(loop.ref_commands),
            kw.LOOPS_FRAMERATE: loop.framerate,
            kw.LOOPS_DELAY: loop.delay,
            kw.LOOPS_TIME_FILTER_NUM: self._handle_image(loop.time_filter_num),
            kw.LOOPS_TIME_FILTER_DEN: self._handle_image(loop.time_filter_den)
        }
        self._add_to_table(kw.LOOPS_TABLE, loop, row)

    def _create_primary_hdu(self) -> fits.PrimaryHDU:
        hdr = fits.Header([(f'HIERARCH {md.key}' if len(md.key) > 8 else md.key,
                            md.value,
                            md.comment) for md in self._system.metadata])
        hdr[kw.AOT_VERSION] = kw.CURRENT_AOT_VERSION
        hdr[kw.AOT_TIMESYS] = kw.AOT_TIMESYS_UTC

        if self._system.ao_mode not in kw.AOT_AO_MODE_SET:
            raise ValueError(f"'AOSystem.ao_mode' must be one of: {str(kw.AOT_AO_MODE_SET)[1:-1]}")
        hdr[kw.AOT_AO_MODE] = self._system.ao_mode

        hdr[kw.AOT_DATE_BEG] = datetime_to_iso(self._system.date_beginning)
        hdr[kw.AOT_DATE_END] = datetime_to_iso(self._system.date_end)

        if self._system.name is not None:
            hdr[kw.AOT_SYSTEM_NAME] = self._system.name
        if self._system.strehl_ratio is not None:
            hdr[kw.AOT_STREHL_RATIO] = self._system.strehl_ratio
        if self._system.temporal_error is not None:
            hdr[kw.AOT_TEMPORAL_ERROR] = self._system.temporal_error
        if self._system.config is not None:
            hdr[kw.AOT_CONFIG] = self._system.config

        return fits.PrimaryHDU(header=hdr)

    def _create_bintable_hdus(self) -> list[fits.BinTableHDU]:
        hdus = []
        for table_name in kw.TABLE_SEQUENCE:
            try:
                table = self._tables[table_name]
            except KeyError:
                if table_name in kw.SECONDARY_TABLE_SET:
                    # If the table is not mandatory and it is not being used, just don't create it
                    continue
                # Otherwise, we need to create an empty table
                table = {}

            columns = []
            for field in kw.TABLE_FIELDS[table_name].values():
                values = [tup[1][field.name] for tup in table.values()]

                if field.format == kw.LIST_FORMAT:
                    array = np.empty(len(values), dtype=np.object_)

                    # Convert every entry to numpy array, if any 64-bit floats are detected we need to use 'D' format.
                    flag = False
                    for i, l in enumerate(values):
                        aux = np.array(l)
                        if aux.dtype != np.float32:
                            aux = aux.astype(np.float64, casting='safe')
                            flag = True
                        array[i] = aux

                    # We always use the 'Q' format, meaning the VLAs use a 64-bit descriptor.
                    # If we used the 'P' format we could potentially save some storage (64-bits per row per VLA column).
                    # However, in scenarios with very large amounts of VLA data, the heap offset could overflow.
                    # This is hard to calculate ahead of time, so we just prefer to take the small bump to file size.
                    # Realistically, this size increase is insignificant when compared to the actual data being stored.
                    col = fits.Column(name=field.name, format=f"Q{'D' if flag else 'E'}", unit=field.unit, array=array)
                else:
                    # Convert to numpy array and try as much as possible to keep the resulting dtype
                    array = np.array(values)
                    if field.format == kw.STRING_FORMAT:
                        col = fits.Column(name=field.name, format=f'{np.char._get_num_chars(array)}A', unit=field.unit,
                                          array=array)
                    elif field.format == kw.INTEGER_FORMAT:
                        if (t := array.dtype) == np.int16:
                            f = '1I'
                        elif t == np.int32:
                            f = '1J'
                        else:
                            if t != np.int64 and array.size > 0:
                                # If not int16, int32 or int64 make one last-ditch effort to convert to int64 by default
                                array = array.astype(np.int64, casting='safe')
                            f = '1K'
                        col = fits.Column(name=field.name, format=f, null=_int_min, unit=field.unit, array=array)
                    elif field.format == kw.FLOAT_FORMAT:
                        if (t := array.dtype) == np.float32:
                            f = '1E'
                        else:
                            if t != np.float64 and array.size > 0:
                                # If not float32 or float64 make one last-ditch effort to convert to float64 by default
                                array = array.astype(np.float64, casting='safe')
                            f = '1D'
                        col = fits.Column(name=field.name, format=f, unit=field.unit, array=array)
                    else:
                        raise NotImplementedError
                columns.append(col)
            hdus.append(fits.BinTableHDU.from_columns(name=table_name, columns=columns))
        return hdus

    def _create_image_hdus(self) -> list[fits.ImageHDU]:
        hdus = []
        for image in self._images.values():
            hdr = fits.Header([(f'HIERARCH {md.key}' if len(md.key) > 8 else md.key,
                                md.value,
                                md.comment) for md in image.metadata])
            if image.time is not None:
                hdr[kw.TIME_REFERENCE] = self._create_row_reference(image.time.uid)
            if image.unit is not None:
                hdr[kw.IMAGE_UNIT] = image.unit
            hdus.append(fits.ImageHDU(name=image.name, data=image.data, header=hdr))
        return hdus



"""
This module contains data that defines the AOT FITS format itself, including the specific FITS keywords and tables used.
Not meant to be imported by users.
"""

import re
from dataclasses import dataclass

CURRENT_AOT_VERSION = '1.0.0'


@dataclass
class AOTField:
    name: str
    format: str
    unit: str
    mandatory: bool = False
    unique: bool = False
    description: str = None


class AOTFieldDict(dict):
    def __init__(self, *args: AOTField):
        super().__init__({field.name: field for field in args})


def fits_type_to_aot(fits_type: str) -> str:
    if fits_type in ['E', 'D']:
        return FLOAT_FORMAT
    if fits_type in ['B', 'I', 'J', 'K']:
        return INTEGER_FORMAT
    if re.fullmatch(r'\d*A', fits_type):
        return STRING_FORMAT
    if re.fullmatch(r'[QP][DE]\(\d*\)', fits_type):
        return LIST_FORMAT
    raise ValueError('Format not in AOT FITS')


STRING_FORMAT = 'str'  # e.g. 64A
INTEGER_FORMAT = 'int'  # e.g. K
FLOAT_FORMAT = 'flt'  # e.g. D
LIST_FORMAT = 'lst'  # e.g. QD

ROW_REFERENCE = 'ROWREF'
INTERNAL_REFERENCE = 'INTREF'
FILE_REFERENCE = 'FILEREF'
URL_REFERENCE = 'URLREF'
IMAGE_UNIT = 'BUNIT'

UNIT_DIMENSIONLESS = '1'
UNIT_COUNT = 'count'
UNIT_METERS = 'm'
UNIT_SECONDS = 's'
UNIT_RADIANS = 'rad'
UNIT_DEGREES = 'deg'
UNIT_ELECTRONS = 'electron'
UNIT_PIXELS = 'pix'
UNIT_DECIBELS = 'dB'
UNIT_FRAME = 'frame'
UNIT_HERTZ = 'Hz'
UNIT_ARCSEC = 'arcsec'

TIME_TABLE = 'AOT_TIME'
ATMOSPHERIC_PARAMETERS_TABLE = 'AOT_ATMOSPHERIC_PARAMETERS'
ABERRATIONS_TABLE = 'AOT_ABERRATIONS'
TELESCOPES_TABLE = 'AOT_TELESCOPES'
SOURCES_TABLE = 'AOT_SOURCES'
SOURCES_SODIUM_LGS_TABLE = 'AOT_SOURCES_SODIUM_LGS'
SOURCES_RAYLEIGH_LGS_TABLE = 'AOT_SOURCES_RAYLEIGH_LGS'
DETECTORS_TABLE = 'AOT_DETECTORS'
SCORING_CAMERAS_TABLE = 'AOT_SCORING_CAMERAS'
WAVEFRONT_SENSORS_TABLE = 'AOT_WAVEFRONT_SENSORS'
WAVEFRONT_SENSORS_SHACK_HARTMANN_TABLE = 'AOT_WAVEFRONT_SENSORS_SHACK_HARTMANN'
WAVEFRONT_SENSORS_PYRAMID_TABLE = 'AOT_WAVEFRONT_SENSORS_PYRAMID'
WAVEFRONT_CORRECTORS_TABLE = 'AOT_WAVEFRONT_CORRECTORS'
WAVEFRONT_CORRECTORS_DM_TABLE = 'AOT_WAVEFRONT_CORRECTORS_DM'
LOOPS_TABLE = 'AOT_LOOPS'
LOOPS_CONTROL_TABLE = 'AOT_LOOPS_CONTROL'
LOOPS_OFFLOAD_TABLE = 'AOT_LOOPS_OFFLOAD'

MANDATORY_TABLE_SET = {TIME_TABLE, ATMOSPHERIC_PARAMETERS_TABLE, ABERRATIONS_TABLE, TELESCOPES_TABLE, SOURCES_TABLE,
                       DETECTORS_TABLE, SCORING_CAMERAS_TABLE, WAVEFRONT_SENSORS_TABLE, WAVEFRONT_CORRECTORS_TABLE,
                       LOOPS_TABLE}

SECONDARY_TABLE_SET = {SOURCES_SODIUM_LGS_TABLE, SOURCES_RAYLEIGH_LGS_TABLE, WAVEFRONT_SENSORS_SHACK_HARTMANN_TABLE,
                       WAVEFRONT_SENSORS_PYRAMID_TABLE, WAVEFRONT_CORRECTORS_DM_TABLE, LOOPS_CONTROL_TABLE,
                       LOOPS_OFFLOAD_TABLE}
TABLE_SET = MANDATORY_TABLE_SET | SECONDARY_TABLE_SET
TABLE_SEQUENCE = [TIME_TABLE, ATMOSPHERIC_PARAMETERS_TABLE, ABERRATIONS_TABLE, TELESCOPES_TABLE, SOURCES_TABLE,
                  SOURCES_SODIUM_LGS_TABLE, SOURCES_RAYLEIGH_LGS_TABLE, DETECTORS_TABLE, SCORING_CAMERAS_TABLE,
                  WAVEFRONT_SENSORS_TABLE, WAVEFRONT_SENSORS_SHACK_HARTMANN_TABLE, WAVEFRONT_SENSORS_PYRAMID_TABLE,
                  WAVEFRONT_CORRECTORS_TABLE, WAVEFRONT_CORRECTORS_DM_TABLE, LOOPS_TABLE, LOOPS_CONTROL_TABLE,
                  LOOPS_OFFLOAD_TABLE]
TABLE_FIELDS: dict[str, AOTFieldDict] = {}

# AOSystem keywords
AOT_VERSION = 'AOT-VERS'
AOT_AO_MODE = 'AO-MODE'
AOT_AO_MODE_SET = {'SCAO', 'SLAO', 'GLAO', 'MOAO', 'LTAO', 'MCAO'}
AOT_TIMESYS = 'TIMESYS'
AOT_TIMESYS_UTC = 'UTC'
AOT_DATE_BEG = 'DATE-BEG'
AOT_DATE_END = 'DATE-END'
AOT_STREHL_RATIO = 'STREHL-R'
AOT_SYSTEM_NAME = 'SYS-NAME'
AOT_TEMPORAL_ERROR = 'TEMP-ERR'
AOT_CONFIG = 'CONFIG'
AOT_HEADER_SET = {AOT_VERSION, AOT_AO_MODE, AOT_TIMESYS, AOT_DATE_BEG, AOT_DATE_END, AOT_STREHL_RATIO, AOT_SYSTEM_NAME,
                  AOT_TEMPORAL_ERROR, AOT_CONFIG}

REFERENCE_UID = 'UID'
TIME_REFERENCE = 'TIME_UID'
ABERRATION_REFERENCE = 'ABERRATION_UID'
LASER_LAUNCH_TELESCOPE_REFERENCE = 'LLT_UID'
DETECTOR_REFERENCE = 'DETECTOR_UID'
SOURCE_REFERENCE = 'SOURCE_UID'
NCPA_REFERENCE = 'NCPA_UID'
TELESCOPE_REFERENCE = 'TELESCOPE_UID'
TRANSFORMATION_MATRIX = 'TRANSFORMATION_MATRIX'

# AOT_TIME fields
TIME_TIMESTAMPS = 'TIMESTAMPS'
TIME_FRAME_NUMBERS = 'FRAME_NUMBERS'
TIME_FIELDS = AOTFieldDict(
    AOTField(name=REFERENCE_UID, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True, unique=True),
    AOTField(name=TIME_TIMESTAMPS, format=LIST_FORMAT, unit=UNIT_SECONDS),
    AOTField(name=TIME_FRAME_NUMBERS, format=LIST_FORMAT, unit=UNIT_COUNT)
)
TABLE_FIELDS[TIME_TABLE] = TIME_FIELDS

# AOT_ATMOSPHERIC_PARAMETERS fields
ATMOSPHERIC_PARAMETERS_WAVELENGTH = 'WAVELENGTH'
ATMOSPHERIC_PARAMETERS_R0 = 'R0'
ATMOSPHERIC_PARAMETERS_FWHM = 'FWHM'
ATMOSPHERIC_PARAMETERS_TAU0 = 'TAU0'
ATMOSPHERIC_PARAMETERS_THETA0 = 'THETA0'
ATMOSPHERIC_PARAMETERS_LAYERS_WEIGHT = 'LAYERS_WEIGHT'
ATMOSPHERIC_PARAMETERS_LAYERS_HEIGHT = 'LAYERS_HEIGHT'
ATMOSPHERIC_PARAMETERS_LAYERS_L0 = 'LAYERS_LO'
ATMOSPHERIC_PARAMETERS_LAYERS_WIND_SPEED = 'LAYERS_WIND_SPEED'
ATMOSPHERIC_PARAMETERS_LAYERS_WIND_DIRECTION = 'LAYERS_WIND_DIRECTION'
ATMOSPHERIC_PARAMETERS_FIELDS = AOTFieldDict(
    AOTField(name=REFERENCE_UID, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True, unique=True),
    AOTField(name=ATMOSPHERIC_PARAMETERS_WAVELENGTH, format=FLOAT_FORMAT, unit=UNIT_METERS),
    AOTField(name=TIME_REFERENCE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=ATMOSPHERIC_PARAMETERS_R0, format=LIST_FORMAT, unit=UNIT_METERS),
    AOTField(name=ATMOSPHERIC_PARAMETERS_FWHM, format=LIST_FORMAT, unit=UNIT_ARCSEC),
    AOTField(name=ATMOSPHERIC_PARAMETERS_TAU0, format=LIST_FORMAT, unit=UNIT_SECONDS),
    AOTField(name=ATMOSPHERIC_PARAMETERS_THETA0, format=LIST_FORMAT, unit=UNIT_RADIANS),
    AOTField(name=ATMOSPHERIC_PARAMETERS_LAYERS_WEIGHT, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=ATMOSPHERIC_PARAMETERS_LAYERS_HEIGHT, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=ATMOSPHERIC_PARAMETERS_LAYERS_L0, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=ATMOSPHERIC_PARAMETERS_LAYERS_WIND_SPEED, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=ATMOSPHERIC_PARAMETERS_LAYERS_WIND_DIRECTION, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=TRANSFORMATION_MATRIX, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS)
)
TABLE_FIELDS[ATMOSPHERIC_PARAMETERS_TABLE] = ATMOSPHERIC_PARAMETERS_FIELDS

# AOT_ABERRATIONS fields
ABERRATION_MODES = 'MODES'
ABERRATION_COEFFICIENTS = 'COEFFICIENTS'
ABERRATION_X_OFFSETS = 'X_OFFSETS'
ABERRATION_Y_OFFSETS = 'Y_OFFSETS'
ABERRATION_FIELDS = AOTFieldDict(
    AOTField(name=REFERENCE_UID, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True, unique=True),
    AOTField(name=ABERRATION_MODES, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True),
    AOTField(name=ABERRATION_COEFFICIENTS, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True),
    AOTField(name=ABERRATION_X_OFFSETS, format=LIST_FORMAT, unit=UNIT_RADIANS),
    AOTField(name=ABERRATION_Y_OFFSETS, format=LIST_FORMAT, unit=UNIT_RADIANS)
)
TABLE_FIELDS[ABERRATIONS_TABLE] = ABERRATION_FIELDS

# AOT_TELESCOPES fields
TELESCOPE_TYPE = 'TYPE'
TELESCOPE_LATITUDE = 'LATITUDE'
TELESCOPE_LONGITUDE = 'LONGITUDE'
TELESCOPE_ELEVATION = 'ELEVATION'
TELESCOPE_AZIMUTH = 'AZIMUTH'
TELESCOPE_PARALLACTIC = 'PARALLACTIC'
TELESCOPE_PUPIL_MASK = 'PUPIL_MASK'
TELESCOPE_PUPIL_ANGLE = 'PUPIL_ANGLE'
TELESCOPE_ENCLOSING_D = 'ENCLOSING_D'
TELESCOPE_INSCRIBED_D = 'INSCRIBED_D'
TELESCOPE_OBSTRUCTION_D = 'OBSTRUCTION_D'
TELESCOPE_SEGMENTS_TYPE = 'SEGMENT_TYPE'
TELESCOPE_SEGMENTS_SIZE = 'SEGMENT_SIZE'
TELESCOPE_SEGMENTS_X = 'SEGMENTS_X'
TELESCOPE_SEGMENTS_Y = 'SEGMENTS_Y'
TELESCOPE_FIELDS = AOTFieldDict(
    AOTField(name=REFERENCE_UID, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True, unique=True),
    AOTField(name=TELESCOPE_TYPE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True),
    AOTField(name=TELESCOPE_LATITUDE, format=FLOAT_FORMAT, unit=UNIT_DEGREES),
    AOTField(name=TELESCOPE_LONGITUDE, format=FLOAT_FORMAT, unit=UNIT_DEGREES),
    AOTField(name=TELESCOPE_ELEVATION, format=FLOAT_FORMAT, unit=UNIT_DEGREES),
    AOTField(name=TELESCOPE_AZIMUTH, format=FLOAT_FORMAT, unit=UNIT_DEGREES),
    AOTField(name=TELESCOPE_PARALLACTIC, format=FLOAT_FORMAT, unit=UNIT_DEGREES),
    AOTField(name=TELESCOPE_PUPIL_MASK, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=TELESCOPE_PUPIL_ANGLE, format=FLOAT_FORMAT, unit=UNIT_RADIANS),
    AOTField(name=TELESCOPE_ENCLOSING_D, format=FLOAT_FORMAT, unit=UNIT_METERS),
    AOTField(name=TELESCOPE_INSCRIBED_D, format=FLOAT_FORMAT, unit=UNIT_METERS),
    AOTField(name=TELESCOPE_OBSTRUCTION_D, format=FLOAT_FORMAT, unit=UNIT_METERS),
    AOTField(name=TELESCOPE_SEGMENTS_TYPE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True),
    AOTField(name=TELESCOPE_SEGMENTS_SIZE, format=FLOAT_FORMAT, unit=UNIT_METERS),
    AOTField(name=TELESCOPE_SEGMENTS_X, format=LIST_FORMAT, unit=UNIT_METERS),
    AOTField(name=TELESCOPE_SEGMENTS_Y, format=LIST_FORMAT, unit=UNIT_METERS),
    AOTField(name=TRANSFORMATION_MATRIX, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=ABERRATION_REFERENCE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS)
)
TABLE_FIELDS[TELESCOPES_TABLE] = TELESCOPE_FIELDS
TELESCOPE_TYPE_MAIN = 'Main Telescope'
TELESCOPE_TYPE_LLT = 'Laser Launch Telescope'
TELESCOPE_SEGMENT_TYPE_MONOLITHIC = 'Monolithic'
TELESCOPE_SEGMENT_TYPE_HEXAGON = 'Hexagon'
TELESCOPE_SEGMENT_TYPE_CIRCLE = 'Circle'

# AOT_SOURCES fields
SOURCE_TYPE = 'TYPE'
SOURCE_RIGHT_ASCENSION = 'RIGHT_ASCENSION'
SOURCE_DECLINATION = 'DECLINATION'
SOURCE_ELEVATION_OFFSET = 'ELEVATION_OFFSET'
SOURCE_AZIMUTH_OFFSET = 'AZIMUTH_OFFSET'
SOURCE_WIDTH = 'WIDTH'
SOURCE_FIELDS = AOTFieldDict(
    AOTField(name=REFERENCE_UID, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True, unique=True),
    AOTField(name=SOURCE_TYPE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True),
    AOTField(name=SOURCE_RIGHT_ASCENSION, format=FLOAT_FORMAT, unit=UNIT_DEGREES),
    AOTField(name=SOURCE_DECLINATION, format=FLOAT_FORMAT, unit=UNIT_DEGREES),
    AOTField(name=SOURCE_ELEVATION_OFFSET, format=FLOAT_FORMAT, unit=UNIT_DEGREES),
    AOTField(name=SOURCE_AZIMUTH_OFFSET, format=FLOAT_FORMAT, unit=UNIT_DEGREES),
    AOTField(name=SOURCE_WIDTH, format=FLOAT_FORMAT, unit=UNIT_RADIANS)
)
TABLE_FIELDS[SOURCES_TABLE] = SOURCE_FIELDS
SOURCE_TYPE_SCIENCE_STAR = 'Science Star'
SOURCE_TYPE_NATURAL_GUIDE_STAR = 'Natural Guide Star'
SOURCE_TYPE_SODIUM_LASER_GUIDE_STAR = 'Sodium Laser Guide Star'
SOURCE_TYPE_RAYLEIGH_LASER_GUIDE_STAR = 'Rayleigh Laser Guide Star'

# AOT_SOURCES_SODIUM_LGS fields
SOURCE_SODIUM_LGS_HEIGHT = 'HEIGHT'
SOURCE_SODIUM_LGS_PROFILE = 'PROFILE'
SOURCE_SODIUM_LGS_ALTITUDES = 'ALTITUDES'
SOURCE_SODIUM_LGS_FIELDS = AOTFieldDict(
    AOTField(name=REFERENCE_UID, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True, unique=True),
    AOTField(name=SOURCE_SODIUM_LGS_HEIGHT, format=FLOAT_FORMAT, unit=UNIT_METERS),
    AOTField(name=SOURCE_SODIUM_LGS_PROFILE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=SOURCE_SODIUM_LGS_ALTITUDES, format=LIST_FORMAT, unit=UNIT_METERS),
    AOTField(name=LASER_LAUNCH_TELESCOPE_REFERENCE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS)
)
TABLE_FIELDS[SOURCES_SODIUM_LGS_TABLE] = SOURCE_SODIUM_LGS_FIELDS

# AOT_SOURCES_RAYLEIGH_LGS fields
SOURCE_RAYLEIGH_LGS_DISTANCE = 'DISTANCE'
SOURCE_RAYLEIGH_LGS_DEPTH = 'DEPTH'
SOURCE_RAYLEIGH_LGS_FIELDS = AOTFieldDict(
    AOTField(name=REFERENCE_UID, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True, unique=True),
    AOTField(name=SOURCE_RAYLEIGH_LGS_DISTANCE, format=FLOAT_FORMAT, unit=UNIT_METERS),
    AOTField(name=SOURCE_RAYLEIGH_LGS_DEPTH, format=FLOAT_FORMAT, unit=UNIT_METERS),
    AOTField(name=LASER_LAUNCH_TELESCOPE_REFERENCE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS)
)
TABLE_FIELDS[SOURCES_RAYLEIGH_LGS_TABLE] = SOURCE_RAYLEIGH_LGS_FIELDS

# AOT_DETECTORS fields
DETECTOR_TYPE = 'TYPE'
DETECTOR_SAMPLING_TECHNIQUE = 'SAMPLING_TECHNIQUE'
DETECTOR_SHUTTER_TYPE = 'SHUTTER_TYPE'
DETECTOR_FLAT_FIELD = 'FLAT_FIELD'
DETECTOR_READOUT_NOISE = 'READOUT_NOISE'
DETECTOR_PIXEL_INTENSITIES = 'PIXEL_INTENSITIES'
DETECTOR_INTEGRATION_TIME = 'INTEGRATION_TIME'
DETECTOR_COADDS = 'COADDS'
DETECTOR_DARK = 'DARK'
DETECTOR_WEIGHT_MAP = 'WEIGHT_MAP'
DETECTOR_QUANTUM_EFFICIENCY = 'QUANTUM_EFFICIENCY'
DETECTOR_PIXEL_SCALE = 'PIXEL_SCALE'
DETECTOR_BINNING = 'BINNING'
DETECTOR_BANDWIDTH = 'BANDWIDTH'
DETECTOR_TRANSMISSION_WAVELENGTH = 'TRANSMISSION_WAVELENGTH'
DETECTOR_TRANSMISSION = 'TRANSMISSION'
DETECTOR_SKY_BACKGROUND = 'SKY_BACKGROUND'
DETECTOR_GAIN = 'GAIN'
DETECTOR_EXCESS_NOISE = 'EXCESS_NOISE'
DETECTOR_FILTER = 'FILTER'
DETECTOR_BAD_PIXEL_MAP = 'BAD_PIXEL_MAP'
DETECTOR_DYNAMIC_RANGE = 'DYNAMIC_RANGE'
DETECTOR_READOUT_RATE = 'READOUT_RATE'
DETECTOR_FRAME_RATE = 'FRAME_RATE'
DETECTOR_FIELDS = AOTFieldDict(
    AOTField(name=REFERENCE_UID, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True, unique=True),
    AOTField(name=DETECTOR_TYPE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=DETECTOR_SAMPLING_TECHNIQUE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=DETECTOR_SHUTTER_TYPE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=DETECTOR_FLAT_FIELD, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=DETECTOR_READOUT_NOISE, format=FLOAT_FORMAT,
             unit=f'{UNIT_ELECTRONS}*{UNIT_SECONDS}^-1*{UNIT_PIXELS}^-1'),
    AOTField(name=DETECTOR_PIXEL_INTENSITIES, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=DETECTOR_INTEGRATION_TIME, format=FLOAT_FORMAT, unit=UNIT_SECONDS),
    AOTField(name=DETECTOR_COADDS, format=INTEGER_FORMAT, unit=UNIT_COUNT),
    AOTField(name=DETECTOR_DARK, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=DETECTOR_WEIGHT_MAP, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=DETECTOR_QUANTUM_EFFICIENCY, format=FLOAT_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=DETECTOR_PIXEL_SCALE, format=FLOAT_FORMAT, unit=f'{UNIT_RADIANS}*{UNIT_PIXELS}^-1'),
    AOTField(name=DETECTOR_BINNING, format=INTEGER_FORMAT, unit=UNIT_COUNT),
    AOTField(name=DETECTOR_BANDWIDTH, format=FLOAT_FORMAT, unit=UNIT_METERS),
    AOTField(name=DETECTOR_TRANSMISSION_WAVELENGTH, format=LIST_FORMAT, unit=UNIT_METERS),
    AOTField(name=DETECTOR_TRANSMISSION, format=LIST_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=DETECTOR_SKY_BACKGROUND, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=DETECTOR_GAIN, format=FLOAT_FORMAT, unit=UNIT_ELECTRONS),
    AOTField(name=DETECTOR_EXCESS_NOISE, format=FLOAT_FORMAT, unit=UNIT_ELECTRONS),
    AOTField(name=DETECTOR_FILTER, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=DETECTOR_BAD_PIXEL_MAP, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=DETECTOR_DYNAMIC_RANGE, format=FLOAT_FORMAT, unit=UNIT_DECIBELS),
    AOTField(name=DETECTOR_READOUT_RATE, format=FLOAT_FORMAT, unit=f'{UNIT_PIXELS}*{UNIT_SECONDS}^-1'),
    AOTField(name=DETECTOR_FRAME_RATE, format=FLOAT_FORMAT, unit=f'{UNIT_FRAME}*{UNIT_SECONDS}^-1'),
    AOTField(name=TRANSFORMATION_MATRIX, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS)
)
TABLE_FIELDS[DETECTORS_TABLE] = DETECTOR_FIELDS

# AOT_SCORING_CAMERAS fields
SCORING_CAMERA_PUPIL_MASK = 'PUPIL_MASK'
SCORING_CAMERA_WAVELENGTH = 'WAVELENGTH'
SCORING_CAMERA_FIELDS = AOTFieldDict(
    AOTField(name=REFERENCE_UID, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True, unique=True),
    AOTField(name=SCORING_CAMERA_PUPIL_MASK, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=SCORING_CAMERA_WAVELENGTH, format=FLOAT_FORMAT, unit=UNIT_METERS),
    AOTField(name=TRANSFORMATION_MATRIX, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=DETECTOR_REFERENCE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=ABERRATION_REFERENCE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS)
)
TABLE_FIELDS[SCORING_CAMERAS_TABLE] = SCORING_CAMERA_FIELDS

# AOT_WAVEFRONT_SENSORS
WAVEFRONT_SENSOR_TYPE = 'TYPE'
WAVEFRONT_SENSOR_DIMENSIONS = 'DIMENSIONS'
WAVEFRONT_SENSOR_N_VALID_SUBAPERTURES = 'N_VALID_SUBAPERTURES'
WAVEFRONT_SENSOR_MEASUREMENTS = 'MEASUREMENTS'
WAVEFRONT_SENSOR_REF_MEASUREMENTS = 'REF_MEASUREMENTS'
WAVEFRONT_SENSOR_SUBAPERTURE_MASK = 'SUBAPERTURE_MASK'
WAVEFRONT_SENSOR_MASK_X_OFFSETS = 'MASK_X_OFFSETS'
WAVEFRONT_SENSOR_MASK_Y_OFFSETS = 'MASK_Y_OFFSETS'
WAVEFRONT_SENSOR_SUBAPERTURE_SIZE = 'SUBAPERTURE_SIZE'
WAVEFRONT_SENSOR_SUBAPERTURE_INTENSITIES = 'SUBAPERTURE_INTENSITIES'
WAVEFRONT_SENSOR_WAVELENGTH = 'WAVELENGTH'
WAVEFRONT_SENSOR_OPTICAL_GAIN = 'OPTICAL_GAIN'
WAVEFRONT_SENSOR_FIELDS = AOTFieldDict(
    AOTField(name=REFERENCE_UID, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True, unique=True),
    AOTField(name=WAVEFRONT_SENSOR_TYPE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True),
    AOTField(name=SOURCE_REFERENCE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True),
    AOTField(name=WAVEFRONT_SENSOR_DIMENSIONS, format=INTEGER_FORMAT, unit=UNIT_COUNT, mandatory=True),
    AOTField(name=WAVEFRONT_SENSOR_N_VALID_SUBAPERTURES, format=INTEGER_FORMAT, unit=UNIT_COUNT, mandatory=True),
    AOTField(name=WAVEFRONT_SENSOR_MEASUREMENTS, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=WAVEFRONT_SENSOR_REF_MEASUREMENTS, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=WAVEFRONT_SENSOR_SUBAPERTURE_MASK, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=WAVEFRONT_SENSOR_MASK_X_OFFSETS, format=LIST_FORMAT, unit=UNIT_PIXELS),
    AOTField(name=WAVEFRONT_SENSOR_MASK_Y_OFFSETS, format=LIST_FORMAT, unit=UNIT_PIXELS),
    AOTField(name=WAVEFRONT_SENSOR_SUBAPERTURE_SIZE, format=FLOAT_FORMAT, unit=UNIT_PIXELS),
    AOTField(name=WAVEFRONT_SENSOR_SUBAPERTURE_INTENSITIES, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=WAVEFRONT_SENSOR_WAVELENGTH, format=FLOAT_FORMAT, unit=UNIT_METERS),
    AOTField(name=WAVEFRONT_SENSOR_OPTICAL_GAIN, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=TRANSFORMATION_MATRIX, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=DETECTOR_REFERENCE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=ABERRATION_REFERENCE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=NCPA_REFERENCE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS)
)
TABLE_FIELDS[WAVEFRONT_SENSORS_TABLE] = WAVEFRONT_SENSOR_FIELDS
WAVEFRONT_SENSOR_TYPE_SHACK_HARTMANN = 'Shack-Hartmann'
WAVEFRONT_SENSOR_TYPE_PYRAMID = 'Pyramid'

# AOT_WAVEFRONT_SENSORS_SHACK_HARTMANN
WAVEFRONT_SENSOR_SHACK_HARTMANN_CENTROIDING_ALGORITHM = 'CENTROIDING_ALGORITHM'
WAVEFRONT_SENSOR_SHACK_HARTMANN_CENTROID_GAINS = 'CENTROID_GAINS'
WAVEFRONT_SENSOR_SHACK_HARTMANN_SPOT_FWHM = 'SPOT_FWHM'
WAVEFRONT_SENSOR_SHACK_HARTMANN_FIELDS = AOTFieldDict(
    AOTField(name=REFERENCE_UID, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True, unique=True),
    AOTField(name=WAVEFRONT_SENSOR_SHACK_HARTMANN_CENTROIDING_ALGORITHM, format=STRING_FORMAT,
             unit=UNIT_DIMENSIONLESS),
    AOTField(name=WAVEFRONT_SENSOR_SHACK_HARTMANN_CENTROID_GAINS, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=WAVEFRONT_SENSOR_SHACK_HARTMANN_SPOT_FWHM, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS)
)
TABLE_FIELDS[WAVEFRONT_SENSORS_SHACK_HARTMANN_TABLE] = WAVEFRONT_SENSOR_SHACK_HARTMANN_FIELDS

# AOT_WAVEFRONT_SENSORS_PYRAMID
WAVEFRONT_SENSOR_PYRAMID_N_SIDES = 'N_SIDES'
WAVEFRONT_SENSOR_PYRAMID_MODULATION = 'MODULATION'
WAVEFRONT_SENSOR_PYRAMID_FIELDS = AOTFieldDict(
    AOTField(name=REFERENCE_UID, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True, unique=True),
    AOTField(name=WAVEFRONT_SENSOR_PYRAMID_N_SIDES, format=INTEGER_FORMAT, unit=UNIT_COUNT, mandatory=True),
    AOTField(name=WAVEFRONT_SENSOR_PYRAMID_MODULATION, format=FLOAT_FORMAT, unit=UNIT_METERS),
)
TABLE_FIELDS[WAVEFRONT_SENSORS_PYRAMID_TABLE] = WAVEFRONT_SENSOR_PYRAMID_FIELDS

# AOT_WAVEFRONT_CORRECTORS fields
WAVEFRONT_CORRECTOR_TYPE = 'TYPE'
WAVEFRONT_CORRECTOR_N_VALID_ACTUATORS = 'N_VALID_ACTUATORS'
WAVEFRONT_CORRECTOR_PUPIL_MASK = 'PUPIL_MASK'
WAVEFRONT_CORRECTOR_TFZ_NUM = 'TFZ_NUM'
WAVEFRONT_CORRECTOR_TFZ_DEN = 'TFZ_DEN'
WAVEFRONT_CORRECTOR_FIELDS = AOTFieldDict(
    AOTField(name=REFERENCE_UID, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True, unique=True),
    AOTField(name=WAVEFRONT_CORRECTOR_TYPE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True),
    AOTField(name=TELESCOPE_REFERENCE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True),
    AOTField(name=WAVEFRONT_CORRECTOR_N_VALID_ACTUATORS, format=INTEGER_FORMAT, unit=UNIT_COUNT),
    AOTField(name=WAVEFRONT_CORRECTOR_PUPIL_MASK, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=WAVEFRONT_CORRECTOR_TFZ_NUM, format=LIST_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=WAVEFRONT_CORRECTOR_TFZ_DEN, format=LIST_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=TRANSFORMATION_MATRIX, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=ABERRATION_REFERENCE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS)
)
TABLE_FIELDS[WAVEFRONT_CORRECTORS_TABLE] = WAVEFRONT_CORRECTOR_FIELDS
WAVEFRONT_CORRECTOR_TYPE_DM = 'Deformable Mirror'
WAVEFRONT_CORRECTOR_TYPE_TTM = 'Tip-Tilt Mirror'
WAVEFRONT_CORRECTOR_TYPE_LS = 'Linear Stage'

# AOT_WAVEFRONT_CORRECTORS_DM fields
WAVEFRONT_CORRECTOR_DM_ACTUATORS_X = 'ACTUATORS_X'
WAVEFRONT_CORRECTOR_DM_ACTUATORS_Y = 'ACTUATORS_Y'
WAVEFRONT_CORRECTOR_DM_INFLUENCE_FUNCTION = 'INFLUENCE_FUNCTION'
WAVEFRONT_CORRECTOR_DM_STROKE = 'STROKE'
WAVEFRONT_CORRECTOR_DM_FIELDS = AOTFieldDict(
    AOTField(name=REFERENCE_UID, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True, unique=True),
    AOTField(name=WAVEFRONT_CORRECTOR_DM_ACTUATORS_X, format=LIST_FORMAT, unit=UNIT_METERS),
    AOTField(name=WAVEFRONT_CORRECTOR_DM_ACTUATORS_Y, format=LIST_FORMAT, unit=UNIT_METERS),
    AOTField(name=WAVEFRONT_CORRECTOR_DM_INFLUENCE_FUNCTION, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=WAVEFRONT_CORRECTOR_DM_STROKE, format=FLOAT_FORMAT, unit=UNIT_METERS)
)
TABLE_FIELDS[WAVEFRONT_CORRECTORS_DM_TABLE] = WAVEFRONT_CORRECTOR_DM_FIELDS

# AOT_LOOPS fields
LOOPS_TYPE = 'TYPE'
LOOPS_COMMANDED = 'COMMANDED_UID'
LOOPS_STATUS = 'STATUS'
LOOPS_COMMANDS = 'COMMANDS'
LOOPS_REF_COMMANDS = 'REF_COMMANDS'
LOOPS_FRAMERATE = 'FRAMERATE'
LOOPS_DELAY = 'DELAY'
LOOPS_TIME_FILTER_NUM = 'TIME_FILTER_NUM'
LOOPS_TIME_FILTER_DEN = 'TIME_FILTER_DEN'
LOOPS_FIELDS = AOTFieldDict(
    AOTField(name=REFERENCE_UID, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True, unique=True),
    AOTField(name=LOOPS_TYPE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True),
    AOTField(name=LOOPS_COMMANDED, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True),
    AOTField(name=TIME_REFERENCE, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=LOOPS_STATUS, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=LOOPS_COMMANDS, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=LOOPS_REF_COMMANDS, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=LOOPS_FRAMERATE, format=FLOAT_FORMAT, unit=UNIT_HERTZ),
    AOTField(name=LOOPS_DELAY, format=FLOAT_FORMAT, unit=UNIT_FRAME),
    AOTField(name=LOOPS_TIME_FILTER_NUM, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=LOOPS_TIME_FILTER_DEN, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS)
)
TABLE_FIELDS[LOOPS_TABLE] = LOOPS_FIELDS
LOOPS_TYPE_CONTROL = 'Control Loop'
LOOPS_TYPE_OFFLOAD = 'Offload Loop'
LOOPS_STATUS_OPEN = 'Open'
LOOPS_STATUS_CLOSED = 'Closed'

# AOT_LOOPS_CONTROL fields
LOOPS_CONTROL_INPUT_SENSOR = 'INPUT_SENSOR_UID'
LOOPS_CONTROL_MODES = 'MODES'
LOOPS_CONTROL_MODAL_COEFFICIENTS = 'MODAL_COEFFICIENTS'
LOOPS_CONTROL_CONTROL_MATRIX = 'CONTROL_MATRIX'
LOOPS_CONTROL_MEASUREMENTS_TO_MODES = 'MEASUREMENTS_TO_MODES'
LOOPS_CONTROL_MODES_TO_COMMANDS = 'MODES_TO_COMMANDS'
LOOPS_CONTROL_INTERACTION_MATRIX = 'INTERACTION_MATRIX'
LOOPS_CONTROL_COMMANDS_TO_MODES = 'COMMANDS_TO_MODES'
LOOPS_CONTROL_MODES_TO_MEASUREMENTS = 'MODES_TO_MEASUREMENTS'
LOOPS_CONTROL_RESIDUAL_COMMANDS = 'RESIDUAL_COMMANDS'
LOOPS_CONTROL_FIELDS = AOTFieldDict(
    AOTField(name=REFERENCE_UID, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True, unique=True),
    AOTField(name=LOOPS_CONTROL_INPUT_SENSOR, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True),
    AOTField(name=LOOPS_CONTROL_MODES, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=LOOPS_CONTROL_MODAL_COEFFICIENTS, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=LOOPS_CONTROL_CONTROL_MATRIX, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=LOOPS_CONTROL_MEASUREMENTS_TO_MODES, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=LOOPS_CONTROL_MODES_TO_COMMANDS, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=LOOPS_CONTROL_INTERACTION_MATRIX, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=LOOPS_CONTROL_COMMANDS_TO_MODES, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=LOOPS_CONTROL_MODES_TO_MEASUREMENTS, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS),
    AOTField(name=LOOPS_CONTROL_RESIDUAL_COMMANDS, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS)
)
TABLE_FIELDS[LOOPS_CONTROL_TABLE] = LOOPS_CONTROL_FIELDS

# AOT_LOOPS_OFFLOAD fields
LOOPS_OFFLOAD_INPUT_CORRECTOR = 'INPUT_CORRECTOR_UID'
LOOPS_OFFLOAD_OFFLOAD_MATRIX = 'OFFLOAD_MATRIX'
LOOPS_OFFLOAD_FIELDS = AOTFieldDict(
    AOTField(name=REFERENCE_UID, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True, unique=True),
    AOTField(name=LOOPS_OFFLOAD_INPUT_CORRECTOR, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS, mandatory=True),
    AOTField(name=LOOPS_OFFLOAD_OFFLOAD_MATRIX, format=STRING_FORMAT, unit=UNIT_DIMENSIONLESS)
)
TABLE_FIELDS[LOOPS_OFFLOAD_TABLE] = LOOPS_OFFLOAD_FIELDS



"""
This module contains classes for describing telescopes in a system. These may be the main telescope itself of laser
launch telescopes.
"""

from dataclasses import dataclass, field

from .aberration import Aberration
from .base import Referenceable, Coordinates

__all__ = ['Segments', 'Monolithic', 'CircularSegments', 'HexagonalSegments', 'Telescope', 'MainTelescope',
           'LaserLaunchTelescope']

from .image import Image


@dataclass(kw_only=True)
class Segments:
    """Abstract class that describes the segments that constitute the pupil."""
    size: float = None
    """Size of the segments that constitute the pupil, measured as the diameter of the smallest circle that contains the
    entirety of a segment. If no segments exist (monolithic) this should be null. (in m units)"""

    coordinates: list[Coordinates] = field(default_factory=list)
    'List of horizontal/vertical coordinates of each segment in the pupil. (in m units)'

    def __post_init__(self):
        if self.__class__ == Segments:
            raise TypeError("Cannot instantiate abstract class.")


class Monolithic(Segments):
    """Describes a monolithic pupil."""

    def __init__(self):
        super().__init__()  # Use default initialization values for Segments


class HexagonalSegments(Segments):
    """Describes the hexagonal segments that constitute the pupil."""
    pass


class CircularSegments(Segments):
    """Describes the circular segments that constitute the pupil."""
    pass


@dataclass(kw_only=True)
class Telescope(Referenceable):
    """Abstract class that contains data related to a telescope in the system."""

    latitude: float = None
    'Latitude of the telescope (World Geodetic System). (in deg units)'

    longitude: float = None
    'Longitude of the telescope (World Geodetic System). (in deg units)'

    elevation: float = None
    """Elevation of the telescope at start, the angle between the object and the observer's local horizon with 0° being
    the horizon and 90° being zenith (horizontal coordinate system). (in deg units)"""

    azimuth: float = None
    """Azimuth of the telescope at start, the angle of the object around the horizon with 0° being North, increasing
    eastward (horizontal coordinate system). (in deg units)"""

    parallactic: float = None
    """Parallactic angle, the spherical angle between the hour circle and the great circle through a celestial object
    and the zenith. (in deg units)"""

    pupil_mask: Image = None
    """Binary image that describes the shape of the pupil. A 1 indicates the presence of the pupil, while a 0 indicates
    the opposite. (Dimensions :math:`h \\times w`, dimensionless quantity, using data type int)"""

    pupil_angle: float = None
    'Clockwise rotation of the pupil mask. (in rad units)'

    enclosing_diameter: float = None
    'Diameter of the smallest circle that contains the entirety of the pupil (enclosing circle). (in m units)'

    inscribed_diameter: float = None
    """Diameter of the largest circle that can be contained in the pupil (inscribed circle).
    On monolithic circular pupils this is equivalent to `enclosing_diameter`. (in m units)"""

    obstruction_diameter: float = None
    'Diameter of the smallest circle that fully contains the central obstruction. (in m units)'

    segments: Segments = field(default_factory=Monolithic)

    transformation_matrix: Image = None
    r"""Matrix that defines 2-dimensional affine transformations over time (:math:`t`) using homogeneous coordinates.
    Any combination of translation, reflection, scale, rotation and shearing can be described via a single
    :math:`3 \times 3` matrix :math:`M` such that :math:`P' = MP`, where :math:`P` is a
    :math:`\begin{bmatrix}x & y & 1 \end{bmatrix}` vector (with :math:`x` and :math:`y` being the original horizontal 
    and vertical coordinates, respectively) and :math:`P'` is a :math:`\begin{bmatrix}x' & y' & 1 \end{bmatrix}`, where
    :math:`x'` and :math:`y'` are the transformed coordinates. All geometry information must be described relative to
    the same reference origin point, from which transformations may occur.
    (Dimensions :math:`3 \times 3 \times t`, dimensionless quantity, using data type flt)"""

    aberration: Aberration = None  # static_map

    def __post_init__(self):
        if self.__class__ == Telescope:
            raise TypeError("Cannot instantiate abstract class.")


class MainTelescope(Telescope):
    """Contains data related to the main telescope in the system. There is only one main telescope in each system."""
    pass


class LaserLaunchTelescope(Telescope):
    """Contains data related to a laser launch telescope in the system. These telescopes produce a laser guide star."""
    pass



"""
This module contains classes that define multidimensional data in AOT and their respective metadata.
"""

from dataclasses import dataclass, field
from typing import Any

import numpy as np

from .base import Metadatum
from .time import Time

__all__ = ['Image']


@dataclass
class Image:
    """Contains multidimensional data and the metadata related to it."""

    name: str
    """Unique name that identifies the data."""

    data: np.ndarray
    """The multi-dimensional data itself."""

    # _: KW_ONLY

    unit: str = None
    """Unit of measurement used for the data."""

    time: Time = None

    metadata: list[Metadatum] = field(default_factory=list)
    """List of Metadatum objects that describe the Image data."""

    def __eq__(self, other):
        return self.name.upper() == other.name.upper() and \
            np.allclose(self.data, other.data) and \
            self.unit == other.unit and \
            self.time == other.time and \
            self.metadata == other.metadata

    def __post_init__(self):
        self.name = self.name.upper()

    def metadata_to_dict(self) -> dict[str, Any]:
        """Return the metadata as a dictionary key->value (comments are ignored)."""
        return {metadatum.key: metadatum.value for metadatum in self.metadata}



"""
Contains the core of aotpy, namely the classes that are necessary to describe different parts of the AO system.

Note that all these functions and objects are available in the main ``aotpy`` namespace.
"""

from .aberration import *
from .ao_system import *
from .atmosphere import *
from .base import *
from .image import *
from .loop import *
from .optical_sensor import *
from .source import *
from .telescope import *
from .time import *
from .wavefront_corrector import *

__all__ = [s for s in dir() if not s.startswith('_')]



"""
This module contains a class for optical aberrations.
"""

from dataclasses import dataclass, field

from .base import Referenceable, Coordinates
from .image import Image

__all__ = ['Aberration']


@dataclass(kw_only=True)
class Aberration(Referenceable):
    """Represents an optical aberration that may exist in a certain part of the AO system."""
    modes: Image
    """Set of :math:`m` different :math:`h \\times w` arrays, each representing the orthonormal basis of the
    corresponding mode. (Dimensions :math:`m \\times h \\times w`, dimensionless quantity, using data type flt)"""

    coefficients: Image
    """Set of :math:`m` coefficients (one for each of the orthonormal basis of modes) for each :math:`n` pupil offset.
    (Dimensions :math:`m \\times n`, in user defined units, using data type flt)"""

    offsets: list[Coordinates] = field(default_factory=list)
    """List of horizontal/vertical offsets from the centre of the field, defined as `Coordinates`. 
    There must be an offset for each set of coefficients."""



"""
This module contains a class for translating data produced by ESO's CIAO system.
"""

import warnings
from datetime import datetime
from pathlib import Path

import numpy as np
from astropy.io import fits

import aotpy
from .eso import ESOTranslator


# TODO set image units


class CIAOTranslator(ESOTranslator):
    """Contains functions for translating telemetry data produced by ESO's CIAO system.

    Parameters
    ----------
    path
        Path to folder containing telemetry data.
    at_number : {1, 2, 3, 4}
        Number of the AT that produced the data.
    """

    def __init__(self, path: str, at_number: int):
        path = Path(path)
        self._at_number = at_number

        with fits.open(path / 'CIAO_LOOP_0001.fits', extname='LoopFrame') as hdus:
            main_hdr = hdus[0].header
            main_loop_frame = hdus['LoopFrame'].data

        self.system = aotpy.AOSystem(
            ao_mode='SCAO',
            name='CIAO',
            strehl_ratio=main_hdr['ESO AOS ATM SR'],
            temporal_error=main_hdr['ESO AOS ATM TERR']
        )
        self.system.main_telescope = aotpy.MainTelescope(
            uid=f'ESO VLT AT{at_number}',
            elevation=main_hdr['ESO TEL ALT'],
            azimuth=self._azimuth_conversion(main_hdr['ESO TEL AZ']),
            parallactic=main_hdr['ESO TEL PRLTIC']
        )

        ngs = aotpy.NaturalGuideStar('NGS',
                                     right_ascension=main_hdr['RA'],
                                     declination=main_hdr['DEC'])

        main_timestamps = main_loop_frame['Seconds'] + main_loop_frame['USeconds'] / 1.e6
        self.system.date_beginning = datetime.utcfromtimestamp(main_timestamps[0])
        self.system.date_end = datetime.utcfromtimestamp(main_timestamps[-1])
        main_frame_numbers = main_loop_frame['FrameCounter']
        loop_time = aotpy.Time('Loop Time', timestamps=main_timestamps.tolist(),
                               frame_numbers=main_frame_numbers.tolist())

        gradients = self._stack_slopes(main_loop_frame['Gradients'], slope_axis=1)
        reference = self._stack_slopes(fits.getdata(path / 'Acq.DET1.REFSLP_0001.fits'), slope_axis=1)[0]
        wfs = aotpy.ShackHartmann(
            uid='WFS',
            source=ngs,
            n_valid_subapertures=68,
            detector=aotpy.Detector('SAPHIRA'),
            measurements=aotpy.Image('Gradients', gradients, time=loop_time),
            ref_measurements=aotpy.Image('Acq.DET1.REFSLP', reference),
            subaperture_intensities=aotpy.Image(f'Intensities', main_loop_frame['Intensities'], time=loop_time),
            centroiding_algorithm=main_hdr['ESO AOS ACQ CENTROID ALGO']
        )

        pix_loop_frame = fits.getdata(path / 'CIAO_PIXELS_0001.fits')
        wfs.detector.pixel_intensities = aotpy.Image(
            'Pixels',
            data=self._get_pixel_data_from_table(pix_loop_frame),
            time=aotpy.Time('Pixel time', frame_numbers=pix_loop_frame['FrameCounter'].tolist())
        )

        ho_dm = aotpy.DeformableMirror('High Order Deformable Mirror (HODM)', telescope=self.system.main_telescope,
                                       n_valid_actuators=60)
        ittm = aotpy.TipTiltMirror('Image Tip-Tilt Mirror (ITTM)', telescope=self.system.main_telescope)

        cm = self._stack_slopes(fits.getdata(path / 'Recn.REC1.CM_0001.fits'), slope_axis=1)
        ho_cm = cm[: ho_dm.n_valid_actuators]
        tt_cm = cm[ho_dm.n_valid_actuators:]

        s2m = self._stack_slopes(fits.getdata(path / 'RecnOptimiser.S2M_0001.fits'), slope_axis=1)
        if main_hdr['ESO AOS CM MODES CONTROLLED'] != s2m.shape[0]:
            warnings.warn("Keyword 'ESO AOS CM MODES CONTROLLED' does not match modes in measurements to modes matrix")
        s2m = aotpy.Image('RecnOptimiser.S2M', s2m)

        m2c = fits.getdata(path / 'RecnOptimiser.M2V_0001.fits')
        ho_m2c = m2c[:ho_dm.n_valid_actuators]
        tt_m2c = m2c[ho_dm.n_valid_actuators:]

        ho_im = self._stack_slopes(fits.getdata(path / 'RecnOptimiser.HO_IM_0001.fits'), slope_axis=0)
        ho_loop = aotpy.ControlLoop(
            'HO Loop',
            input_sensor=wfs,
            commanded_corrector=ho_dm,
            time=loop_time,
            commands=aotpy.Image('HODM positions', main_loop_frame['HODM_Positions'], time=loop_time),
            ref_commands=aotpy.Image('HOCtr.ACT_POS_REF_MAP',
                                     fits.getdata(path / 'HOCtr.ACT_POS_REF_MAP_0001.fits')[0]),
            control_matrix=aotpy.Image('HO Control Matrix', ho_cm),
            measurements_to_modes=s2m,
            modes_to_commands=aotpy.Image('HO modes to commands', ho_m2c),
            interaction_matrix=aotpy.Image('RecnOptimiser.HO_IM', ho_im),
            framerate=main_hdr['ESO AOS LOOP RATE'],
            closed=main_hdr['ESO AOS HO LOOP ST']
        )

        tt_loop = aotpy.ControlLoop(
            'TT Loop',
            input_sensor=wfs,
            commanded_corrector=ittm,
            time=loop_time,
            commands=aotpy.Image('ITTM positions', main_loop_frame['ITTM_Positions'], time=loop_time),
            ref_commands=aotpy.Image('ESO AOS TTM REFPOS',
                                     np.array([main_hdr['ESO AOS TTM REFPOS X'], main_hdr['ESO AOS TTM REFPOS Y']])),
            control_matrix=aotpy.Image('TT Control Matrix', tt_cm),
            measurements_to_modes=s2m,
            modes_to_commands=aotpy.Image('TT modes to commands', tt_m2c),
            framerate=main_hdr['ESO AOS LOOP RATE'],
            closed=main_hdr['ESO AOS TT LOOP ST']
        )

        asm = aotpy.AtmosphericParameters(
            'ESO ASM (Astronomical Site Monitor)',
            wavelength=500e-9,
            fwhm=[main_hdr['ESO TEL AMBI FWHM']],
            tau0=[main_hdr['ESO TEL AMBI TAU0']],
            theta0=[main_hdr['ESO TEL AMBI THETA0']],
            layers_wind_direction=aotpy.Image('ESO TEL AMBI WINDDIR', np.array([[main_hdr['ESO TEL AMBI WINDDIR']]])),
            layers_wind_speed=aotpy.Image('ESO TEL AMBI WINDSP', np.array([[main_hdr['ESO TEL AMBI WINDSP']]]))
        )

        aos = aotpy.AtmosphericParameters(
            'AO System',
            fwhm=[main_hdr['ESO AOS ATM SEEING']],
            tau0=[main_hdr['ESO AOS ATM TAU0']],
        )

        self.system.sources = [ngs]
        self.system.wavefront_sensors = [wfs]
        self.system.wavefront_correctors = [ho_dm, ittm]
        self.system.loops = [ho_loop, tt_loop]
        self.system.atmosphere_params = [asm, aos]

    def _get_eso_telescope_name(self) -> str:
        # Allow for both:
        #   ESO-VLTI-Amnop
        #   ESO-VLTI-Uijkl-Amnop
        # as long as mnop contains the correct number
        return f"ESO-VLTI-%A%{self._at_number}%"

    def _get_eso_ao_name(self) -> str:
        return 'CIAO'

    def _get_run_id(self) -> str:
        return '60.A-9278(C)'



"""
This module contains classes that describe conceptual loops in an adaptive optics system.
"""

from dataclasses import dataclass

from .base import Referenceable
from .image import Image
from .optical_sensor import WavefrontSensor
from .time import Time
from .wavefront_corrector import WavefrontCorrector

__all__ = ['Loop', 'ControlLoop', 'OffloadLoop']


@dataclass(kw_only=True)
class Loop(Referenceable):
    """Base class that contains data regarding one system loop."""

    commanded_corrector: WavefrontCorrector
    """Wavefront corrector that is being commanded by this loop."""

    time: Time = None

    closed: bool = True
    "Indicates whether the loop was opened or closed for the duration of the data collection."

    commands: Image = None

    """Sequence of commands sent to the associated wavefront corrector. Each of the :math:`t` frames contains the values
    sent to each of the :math:`a_v` valid actuators of a certain wavefront corrector.
    (Dimensions :math:`t \\times a_v`, in m units, using data type flt)"""

    ref_commands: Image = None
    """Reference offset commands for each of :math:`a_v` actuators.
    (Dimensions :math:`a_v`, in m units, using data type flt)"""

    framerate: float = None
    'Frequency at which the loop operates. (in Hz units)'

    delay: float = None
    """Full AO loop delay, measured from the mid-point of the integration to the mid-point of the command applied.
    Includes the RTC latency and other communication delays. (in frame units)"""

    time_filter_num: Image = None
    """One set of time filter numerators per mode being described. If :math:`m=1`, it is assumed to be applicable to all
    modes. The first numerator is the loop gain.
    (Dimensions :math:`m \\times i`, dimensionless quantity, using data type flt)"""

    time_filter_den: Image = None
    r"""One set of time filter denominators per mode being described. If :math:`m=1`, it is assumed to be applicable to
    all modes. Uses standard transfer function :math:`sys(z) = \sum_{i=0}^{N-1} b_i z^i/\sum_{j=0}^N a_j z^j`.
    (Dimensions :math:`m \times j`, dimensionless quantity, using data type flt)"""


@dataclass(kw_only=True)
class ControlLoop(Loop):
    """Contains data relevant to a control loop (relation between one wavefront sensor and one wavefront corrector)."""

    input_sensor: WavefrontSensor

    modes: Image = None
    """Set of :math:`m` different :math:`h \\times w` arrays, each representing the orthonormal basis of the
    corresponding mode. (Dimensions :math:`m \\times h \\times w`, dimensionless quantity, using data type flt)"""

    modal_coefficients: Image = None
    """Sequence of coefficients of the modes to be corrected on the wavefront corrector. Each of the :math:`t` frames
    contains the coefficients respective to each of the :math:`m` modes being corrected.
    (Dimensions :math:`t \\times m`, in user defined units, using data type flt)"""

    control_matrix: Image = None
    """Linear relationship between the wavefront sensor measurements (:math:`d \\times s_v`) and the corrector commands 
    (:math:`a_v`). (Dimensions :math:`a_v \\times d \\times s_v`, in user defined units, using data type flt)"""

    measurements_to_modes: Image = None
    """Linear relationship between the wavefront sensor measurements (:math:`d \\times s_v`) and the modes to be
    corrected (:math:`m`). (Dimensions :math:`m \\times d \\times s_v`, in user defined units, using data type flt)"""

    modes_to_commands: Image = None
    """Linear relationship between the modes (:math:`m`) to be corrected and the corrector commands (:math:`a_v`).
    (Dimensions :math:`a_v \\times m`, in user defined units, using data type flt)"""

    interaction_matrix: Image = None
    """Represents the measurements of the wavefront sensor (:math:`s_v \\times d`) in response to each actuator of the
    wavefront corrector (:math:`a_v`).
    (Dimensions :math:`s_v \\times d \\times a_v`, in user defined units, using data type flt)"""

    commands_to_modes: Image = None
    """Represents the modal response (:math:`m`) to each actuator of the wavefront corrector (:math:`a_v`).
    (Dimensions :math:`m \\times a_v`, in user defined units, using data type flt)"""

    modes_to_measurements: Image = None
    """Linear relationship between the modal response (:math:`m`) and the wavefront sensor measurements 
    (:math:`s_v \\times d`). (Dimensions :math:`s_v \\times d \\times m`, in user defined units, using data type flt)"""

    residual_commands: Image = None
    """Reconstructed corrector commands before time filtering
    (Dimensions :math:`t \\times a_v`, in user defined units, using data type flt)"""


@dataclass(kw_only=True)
class OffloadLoop(Loop):
    """Contains data relevant to an offload loop (relation between two wavefront correctors)."""

    input_corrector: WavefrontCorrector

    offload_matrix: Image = None
    """Linear relationship between the commands from one actuator (:math:`a1`) and the corresponding commands that get
    offloaded to another actuator (:math:`a2`).
    (Dimensions :math:`a2_v \\times a1_v`, in user defined units, using data type flt)"""



"""
aotpy is a package that supports the Adaptive Optics Telemetry (AOT) data exchange standard.
"""
from typing import Type

# Initialize dictionaries of available writers/readers in order to enable their usage in AOSystem
_AVAILABLE_WRITERS: dict[str, Type['SystemWriter']] = {}
_AVAILABLE_READERS: dict[str, Type['SystemReader']] = {}

# First import core and then import io to force dictionaries to be updated
from .core import *
from .io import *

__all__ = [s for s in dir() if not s.startswith('_')]



"""
This module contains a class that defines an adaptive optics system.
"""

import os
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path

from .atmosphere import AtmosphericParameters
from .base import Metadatum
from .loop import Loop
from .optical_sensor import WavefrontSensor, ScoringCamera
from .source import Source
from .telescope import MainTelescope
from .wavefront_corrector import WavefrontCorrector
from .. import _AVAILABLE_WRITERS, _AVAILABLE_READERS

__all__ = ['AOSystem']


@dataclass(kw_only=True)
class AOSystem:
    """Contains all the relevant information about an adaptive optics system.
    Different parts of the system can be accessed via the lists of objects it contains."""

    ao_mode: str = None
    """Describes the system's AO configuration.
    
    Must be one of ``'SCAO'``, ``'SLAO'``, ``'GLAO'``, ``'MOAO'``, ``'LTAO'`` or ``'MCAO'`` (which respectively stand
    for Single Conjugate, Single Laser, Ground Layer, Multi-Object, Laser Tomography and Multi-Conjugate Adaptive
    Optics)."""

    date_beginning: datetime = None
    'Start time of data acquisition.'

    date_end: datetime = None
    'Stop time of data acquisition.'

    name: str = None
    'Name of the AO system that produced the data.'

    strehl_ratio: float = None
    'Estimated strehl ratio (arcsec).'

    temporal_error: float = None
    'Estimated temporal error TODO: (units?).'

    config: str = None
    'Free-form text that describes configuration parameters of the system.'

    atmosphere_params: list[AtmosphericParameters] = field(default_factory=list)
    """List of all atmospheric data related to the system."""

    main_telescope: MainTelescope = None
    """The main telescope of the system."""

    sources: list[Source] = field(default_factory=list)
    """List of all the light sources in the system."""

    scoring_cameras: list[ScoringCamera] = field(default_factory=list)
    """List of all the scoring cameras in the system."""

    wavefront_sensors: list[WavefrontSensor] = field(default_factory=list)
    """List of all wavefront sensors in the system."""

    wavefront_correctors: list[WavefrontCorrector] = field(default_factory=list)
    """List of all wavefront correctors in the system."""

    loops: list[Loop] = field(default_factory=list)
    """List of all loops in the system."""

    metadata: list[Metadatum] = field(default_factory=list)
    """List of Metadatum objects that are associated with the overall system."""

    def write_to_file(self, filename: str | os.PathLike, **kwargs) -> None:
        """
        Writes `AOSystem` to a file. The writing function is deduced by the extension in the specified `filename`.

        Parameters
        ----------
        filename
            Path to the file to be written.
        kwargs
            Optional keyword arguments passed on as options to the writer function.
        """
        ext = Path(filename).suffix[1:]
        if (e := ext.lower()) in _AVAILABLE_WRITERS:
            _AVAILABLE_WRITERS[e](self).write(filename, **kwargs)
        else:
            raise ValueError(f"No available writer for extension '{ext}'. "
                             f"Available extensions: {str(list(_AVAILABLE_WRITERS.keys()))[1:-1]}")

    @staticmethod
    def read_from_file(filename: str | os.PathLike, **kwargs) -> 'AOSystem':
        """
        Reads `AOSystem` from a file. The reading function is deduced by the extension in the specified `filename`.

        Parameters
        ----------
        filename
            Path to the file to be read.
        kwargs
            Optional keyword arguments passed on as options to the reader function.
        """
        ext = Path(filename).suffix[1:]
        if (e := ext.lower()) in _AVAILABLE_READERS:
            return _AVAILABLE_READERS[e](filename, **kwargs).get_system()
        else:
            raise ValueError(f"No available reader for extension '{ext}'. "
                             f"Available extensions: {str(list(_AVAILABLE_READERS.keys()))[1:-1]}")

    def __str__(self) -> str:
        if self.name:
            out = self.name
        else:
            out = 'Unnamed AOSystem'
        out += " telemetry data recording "

        if self.date_beginning is not None:
            out += f"started at {self.date_beginning} "
        if self.date_end is not None:
            out += f"ended at {self.date_beginning} "
        out += f"(configuration '{self.ao_mode}')."
        out += f"\n\tMain Telescope: '{self.main_telescope.uid}'"
        aux = (
            (self.atmosphere_params, 'Atmospheric Parameters'),
            (self.sources, 'Light Sources'),
            (self.scoring_cameras, 'Scoring Cameras'),
            (self.wavefront_sensors, 'Wavefront Sensors'),
            (self.wavefront_correctors, 'Wavefront Correctors'),
            (self.loops, 'System Loops')
        )
        for refs, name in aux:
            if refs:
                out += f'\n\t{name} ({len(refs)}):'
                for i, ref in enumerate(refs):
                    out += f"\n\t\t{i}: '{ref.uid}' ({type(ref).__name__})"

        return out



"""
This module contains a class for describing atmospheric parameters.
"""

from dataclasses import dataclass, field

from .base import Referenceable
from .image import Image
from .time import Time

__all__ = ['AtmosphericParameters']


@dataclass(kw_only=True)
class AtmosphericParameters(Referenceable):
    """Contains atmospheric data relevant to the telemetry data recorded."""

    wavelength: float = None
    'Reference wavelength. (in m units)'

    time: Time = None

    r0: list[float] = field(default_factory=list)
    'List of Fried parameters at reference wavelength at zenith, over time. (in m units)'

    fwhm: list[float] = field(default_factory=list)
    'List of full width at half maximum measures of the seeing disc at zenith, over time. (in arcsec units)'

    tau0: list[float] = field(default_factory=list)
    'List of atmospheric coherence times, over time. (in s units)'

    theta0: list[float] = field(default_factory=list)
    'List of isoplanatic angles, over time. (in rad units)'

    layers_weight: Image = None
    """Fractional weight of each :math:`l` turbulence layer (sums to 1), for each time instant.
    (Dimensions :math:`t \\times l`, dimensionless quantity, using data type flt)"""

    layers_height: Image = None
    """Height above observatory at zenith of each :math:`l` turbulence layer, for each time instant.
    (Dimensions :math:`t \\times l`, in m units, using data type flt)"""

    layers_l0: Image = None
    """Outer scale of turbulence at reference wavelength at zenith of each :math:`l` turbulence layer, for each time
    instant. (Dimensions :math:`t \\times l`, in m units, using data type flt)"""

    layers_wind_speed: Image = None
    r"""Wind speed of each :math:`l` turbulence layer, for each time instant.
    (Dimensions :math:`t \times l`, in ms\ :math:`^{-1}` units, using data type flt)"""

    layers_wind_direction: Image = None
    """Wind direction of each :math:`l` turbulence layer, for each time instant, with 0° being North, increasing
    eastward. (Dimensions :math:`t \\times l`, in deg units, using data type flt)"""

    transformation_matrix: Image = None
    r"""Matrix that defines 2-dimensional affine transformations over time (:math:`t`) using homogeneous coordinates.
    Any combination of translation, reflection, scale, rotation and shearing can be described via a single
    :math:`3 \times 3` matrix :math:`M` such that :math:`P' = MP`, where :math:`P` is a
    :math:`\begin{bmatrix}x & y & 1 \end{bmatrix}` vector (with :math:`x` and :math:`y` being the original horizontal 
    and vertical coordinates, respectively) and :math:`P'` is a :math:`\begin{bmatrix}x' & y' & 1 \end{bmatrix}`, where
    :math:`x'` and :math:`y'` are the transformed coordinates. All geometry information must be described relative to
    the same reference origin point, from which transformations may occur.
    (Dimensions :math:`3 \times 3 \times t`, dimensionless quantity, using data type flt)"""



"""
This module contains classes and functions that aid in handling AOT FITS files.
"""

import datetime
import os
import re

import numpy as np
from astropy.io import fits

import aotpy
from . import _keywords as kw

__all__ = ['FITSFileImage', 'FITSURLImage', 'image_from_file', 'image_from_hdus', 'image_from_hdu',
           'metadatum_from_card', 'metadata_from_hdu', 'datetime_to_iso', 'keyword_is_relevant']


def keyword_is_relevant(keyword):
    """Check if keyword is relevant. Keywords are considered "irrelevant" if they are already reflected elsewhere in the
     object produced by Astropy."""
    _standard_keywords = {'SIMPLE', 'EXTEND', 'BSCALE', 'BZERO', 'XTENSION', 'BITPIX',
                          'PCOUNT', 'GCOUNT', 'EXTNAME', 'CHECKSUM', 'DATASUM'}
    _standard_patterns = re.compile(r'NAXIS\d*')

    return keyword not in _standard_keywords and not _standard_patterns.fullmatch(keyword)


class _FITSExternalImage(aotpy.Image):
    """Describes an external FITS file containing multidimensional data and metadata related to it."""

    def to_internal(self) -> aotpy.Image:
        """
        Convert external image into an internal image.
        """
        return aotpy.Image(name=self.name,
                           data=self.data,
                           unit=self.unit,
                           time=self.time,
                           metadata=self.metadata)


class FITSFileImage(_FITSExternalImage):
    """Describes an external FITS file present locally, containing multidimensional data and metadata related to it.

    `FITSFileImage` remembers the path from which the data came from, enabling future referencing of that path.

    Parameters
    ----------
    path
        Path to FITS file that contains multidimensional data.
    index: int, optional
        Index of the HDU that contains the image data. If omitted, the first HDU containing image data is assumed.
    **kwargs
        Keyword arguments passed on as options to the file handling function.
    """

    def __init__(self, path: str | os.PathLike, index: int = None, **kwargs):
        self.filename = os.path.basename(path)
        self.index = index
        self.name, self.data, self.unit, self._time, self.metadata = _get_image_fields_from_file(path, index, **kwargs)

    def __eq__(self, other):
        return self.filename == other.filename and self.index == other.index and self.time == other.time


class FITSURLImage(_FITSExternalImage):
    """Describes an external FITS file available via URL, containing multidimensional data and metadata related to it.

    `FITSURLImage` remembers the URL from which the data came from, enabling future referencing of that URL.

    Parameters
    ----------
    url
        URL to FITS file that contains multidimensional data.
    index: int, optional
        Index of the HDU that contains the image data. If omitted, the first HDU containing image data is assumed.
    **kwargs
        Keyword arguments passed on as options to the file handling function.
    """

    def __init__(self, url: str, index: int = None, **kwargs):
        self.url = url
        self.index = index
        self.name, self.data, self.unit, self._time, self.metadata = _get_image_fields_from_file(url, index, **kwargs)

    def __eq__(self, other):
        return self.url == other.url and self.index == other.index and self.time == other.time


def image_from_file(path: str | os.PathLike, index: int = None, *, name: str = None, **kwargs) -> aotpy.Image:
    """
    Get `Image` from specified path or URL.

    Parameters
    ----------
    path
        Path/URL to FITS file that contains multidimensional data.
    index: int, optional
        Index of the HDU that contains the image data. If omitted, the first HDU containing image data is assumed.
    name: str, optional
        Name of the Image. If None, the name is the same as in the file.
    **kwargs
        Keyword arguments passed on as options to the file handling function.
    """
    _name, data, unit, _time, metadata = _get_image_fields_from_file(path, index, **kwargs)
    if name is None:
        name = _name
    image = aotpy.Image(name=name, data=data, unit=unit, metadata=metadata)
    image._time = _time
    return image


def image_from_hdus(hdus: fits.HDUList, index: int = None, *, name: str = None) -> aotpy.Image:
    """
    Get `Image` from specified path or URL.

    Parameters
    ----------
    hdus
        List of HDUs from which `Image` is to be extracted.
    index: int, optional
        Index of the HDU that contains the image data. If omitted, the first HDU containing image data is assumed.
    name: str, optional
        Name of the Image. If None, the name is the same as in the HDU.
    """
    _name, data, unit, _time, metadata = _get_image_fields_from_hdus(hdus, index)
    if name is None:
        name = _name
    image = aotpy.Image(name=name, data=data, unit=unit, metadata=metadata)
    image._time = _time
    return image


def image_from_hdu(hdu: fits.ImageHDU, *, name: str = None) -> aotpy.Image:
    """
    Get `Image` from specified HDU.

    Parameters
    ----------
    hdu
        HDU from which `Image` is to be extracted.
    name: str, optional
        Name of the Image. If None, the name is the same as in the HDU.
    """
    _name, data, unit, _time, metadata = _get_image_fields_from_hdu(hdu)
    if name is None:
        name = _name
    image = aotpy.Image(name=name, data=data, unit=unit, metadata=metadata)
    image._time = _time
    return image


def _get_image_fields_from_file(path: str | os.PathLike, index: int = None, **kwargs) -> \
        tuple[str, np.ndarray, str, str, list[aotpy.Metadatum]]:
    try:
        with fits.open(path, **kwargs) as hdus:
            return _get_image_fields_from_hdus(hdus, index)
    except FileNotFoundError:
        from tkinter.filedialog import askopenfilename
        filename = os.path.basename(path)
        selected = askopenfilename(title=f"Please select '{filename}'.",
                                   initialfile=filename, initialdir=os.path.dirname(path),
                                   filetypes=(('FITS files', '*.fits'),
                                              ('Compressed FITS files', '*.fits.gz')))
        if not selected:
            raise ValueError(f"Could not find '{filename}' automatically. Please select it manually")
        with fits.open(selected, **kwargs) as hdus:
            return _get_image_fields_from_hdus(hdus, index)


def _get_image_fields_from_hdus(hdus: fits.HDUList, index: int = None) -> \
        tuple[str, np.ndarray, str, str, list[aotpy.Metadatum]]:
    if index is None:
        for hdu in hdus:
            if hdu.is_image and hdu.data is not None:
                break
        else:
            raise ValueError('Could not find any image data in FITS file.')
    else:
        hdu = hdus[index]
        if not hdu.is_image or hdu.data is None:
            raise ValueError(f'Referenced HDU {hdu.name} does not contain image data.')

    return _get_image_fields_from_hdu(hdu)


def _get_image_fields_from_hdu(hdu) -> tuple[str, np.ndarray, str, str, list[aotpy.Metadatum]]:
    metadata = metadata_from_hdu(hdu)
    unit = None
    if (md := next((x for x in metadata if x.key == kw.IMAGE_UNIT), None)) is not None:
        unit = md.value

    _time = None
    if (md := next((x for x in metadata if x.key == kw.TIME_REFERENCE), None)) is not None:
        _time = md.value
        metadata.remove(md)
    return hdu.name, hdu.data, unit, _time, metadata


def metadatum_from_card(card: fits.Card):
    """
    Get `Metadatum` from `Card`.

    Parameters
    ----------
    card
        `Card` to convert to `Metadatum`.
    """
    return aotpy.Metadatum(card.keyword, card.value, card.comment if card.comment != '' else None)


def metadata_from_hdu(hdu: fits.ImageHDU) -> list[aotpy.Metadatum]:
    """
    Get metadata (list of `Metadatum`) from HDU. Only relevant keywords are extracted.

    Parameters
    ----------
    hdu
        HDU from which to extract metadata.
    """
    # If the keywords are irrelevant they don't need to be part of the image metadata, as that information is already
    # implied in the numpy data.
    return [metadatum_from_card(card) for card in hdu.header.cards if keyword_is_relevant(card.keyword)]


def datetime_to_iso(d: datetime.datetime):
    """
    Convert datetime to ISO format string. If timezone information is present, convert time to UTC and remove
    information.

    Parameters
    ----------
    d
        `datetime` to be converted to an ISO string.
    """
    if d is None:
        return ''
    if d.tzinfo is not None:
        # If datetime has timezone data, convert the datetime to UTC and then remove the timezone data.
        # This is done to ensure dt.isoformat() doesn't print UTC offsets.
        # If datetime doesn't have timezone data, we assume it is already in UTC.
        d = d.astimezone(datetime.timezone.utc).replace(tzinfo=None)
    return d.isoformat()



"""
This module contains a base class for translating non-standard AO telemetry data.
"""

from abc import ABC, abstractmethod

import aotpy


class BaseTranslator(ABC):
    """Abstract class for translators.

    Translators are able to convert non-standard AO telemetry data files into an `AOSystem` object.

    Parameters
    ----------
    *args
        Arguments used to initialize the translator class.

    """

    @abstractmethod
    def __init__(self, *args) -> None:
        self.system: aotpy.AOSystem = aotpy.AOSystem()

    @classmethod
    def translate(cls, *args) -> aotpy.AOSystem:
        """
        Initialize class with `args`, return translated `AOSystem`.

        Parameters
        ----------
        *args
            Arguments used to initialize the translator class.

        Returns
        -------
            `AOSystem` containing translated data.
        """
        t = cls(*args)
        return t.system



"""
This subpackage contains modules and packages for handling AOT files using data storage formats supported by aotpy.
Currently only FITS is supported.
"""

from .base import SystemReader, SystemWriter
from .fits import *
from .. import _AVAILABLE_WRITERS, _AVAILABLE_READERS

# Add available writers/readers to their respective dictionaries
_AVAILABLE_WRITERS['fits'] = FITSWriter
_AVAILABLE_READERS['fits'] = FITSReader



"""
This module contains classes that describe different types of light sources in an AO system.
"""

from dataclasses import dataclass, field

from .base import Referenceable
from .image import Image
from .telescope import LaserLaunchTelescope

__all__ = ['Source', 'ScienceStar', 'NaturalGuideStar', 'SodiumLaserGuideStar', 'RayleighLaserGuideStar']


@dataclass(kw_only=True)
class Source(Referenceable):
    """Abstract class that contains data regarding a light source used by the system."""

    right_ascension: float = None
    'Right ascension of the light source, epoch J2000 (equatorial coordinate system). (in deg units)'

    declination: float = None
    'Declination of the light source, epoch J2000 (equatorial coordinate system). (in deg units)'

    elevation_offset: float = None
    "Offset from the Main Telescope's `elevation`. (in deg units)"

    azimuth_offset: float = None
    "Offset from the Main Telescope's `azimuth`. (in deg units)"

    width: float = None
    'Effective width at zenith. (in rad units)'

    def __post_init__(self):
        if self.__class__ == Source:
            raise TypeError("Cannot instantiate abstract class.")


@dataclass(kw_only=True)
class ScienceStar(Source):
    """Contains data regarding a science star being observed by the system."""
    pass


@dataclass(kw_only=True)
class NaturalGuideStar(Source):
    """Contains data regarding a natural guide star being observed by the system."""
    pass


@dataclass(kw_only=True)
class LaserGuideStar(Source):
    """Abstract class that contains data regarding a laser guide star being observed by the system."""

    laser_launch_telescope: LaserLaunchTelescope = None

    def __post_init__(self):
        if self.__class__ == LaserGuideStar:
            raise TypeError("Cannot instantiate abstract class.")


@dataclass(kw_only=True)
class SodiumLaserGuideStar(LaserGuideStar):
    """Contains data regarding a sodium laser guide star being observed by the system."""

    height: float = None
    'Mean LGS height at zenith. (in m units)'

    profile: Image = None
    """Normalised LGS profile (each set of :math:`l` layers :math:`\\sum = 1`) at zenith, over time.
    (Dimensions :math:`t \\times l`, dimensionless quantity, using data type flt)"""

    altitudes: list[float] = field(default_factory=list)
    """LGS layer profile altitudes at zenith. Must be the same length as the number of layers defined in `profile`.
    (in m units)"""


@dataclass(kw_only=True)
class RayleighLaserGuideStar(LaserGuideStar):
    """Contains data regarding a Rayleigh laser guide star being observed by the system."""

    distance: float = None
    'Fixed distance of the LGS from the telescope. (in m units)'

    depth: float = None
    'TODO: Depth (in m units)'
